import re
from collections import defaultdict
import numpy as np
from numbers import Number
import copy

from netket.jax import canonicalize_dtypes
from netket.utils.types import DType, PyTree

OperatorTuple = tuple[int, int]
r""" Creation and annihilation operators at mode i are encoded as
:math:`\hat{a}_i^\dagger`: (i, 1)
:math:`\hat{a}`: (i, 0)
"""

OperatorTerm = tuple[OperatorTuple, ...]
r""" A term of the form :math:`\hat{a}_1^\dagger \hat{a}_2` would take the form
`((1,1), (2,0))`, where (1,1) represents :math:`\hat{a}_1^\dagger` and (2,0)
represents :math:`\hat{a}_2`."""

OperatorTermsList = list[OperatorTerm]
""" A list of operators that would e.g. describe a Hamiltonian """

OperatorWeightsList = list[Number]
""" A list of weights of corresponding terms """

OperatorDict = dict[OperatorTerm, Number]
""" A dict containing OperatorTerm as key and weights as the values """


def _normal_order_term(
    term: OperatorTerm, weight: Number = 1.0
) -> tuple[OperatorTermsList, OperatorWeightsList]:
    """
    Return a normal ordered single term of the fermion operator.
    Normal ordering corresponds to placing creating operators on the left
    and annihilation on the right.
    Then, it places the highest index on the left and lowest index on the right
    In this ordering, we make sure to account for the anti-commutation of operators.
    """

    parity = -1
    term = copy.deepcopy(list(term))
    weight = copy.copy(weight)

    if len(term) == 0:  # a constant
        return [term], [weight]
    ordered_terms = []
    ordered_weights = []
    # the arguments given to this function will be transformed in a normal ordered way
    # loop through all the operators in the single term from left to right and order them
    # by swapping the term operators (and transform the weights by multiplying with the parity factors)
    for i in range(1, len(term)):
        for j in range(i, 0, -1):
            right_term = term[j]
            left_term = term[j - 1]

            # exchange operators if creation operator is on the right and annihilation on the left
            if right_term[1] and not left_term[1]:
                term[j - 1] = right_term
                term[j] = left_term
                weight *= parity

                # if same indices switch order (creation on the left), remember a a^ = 1 + a^ a
                if right_term[0] == left_term[0]:
                    new_term = term[: (j - 1)] + term[(j + 1) :]

                    # ad the processed term
                    o, w = _normal_order_term(tuple(new_term), parity * weight)
                    ordered_terms += o
                    ordered_weights += w

            # if we have two creation or two annihilation operators
            elif right_term[1] == left_term[1]:
                # If same two Fermionic operators are repeated,
                # evaluate to zero.
                if parity == -1 and right_term[0] == left_term[0]:
                    return ordered_terms, ordered_weights

                # swap if same type but order is not correct
                elif right_term[0] > left_term[0]:
                    term[j - 1] = right_term
                    term[j] = left_term
                    weight *= parity

    ordered_terms.append(term)
    ordered_weights.append(weight)
    return ordered_terms, ordered_weights


def _normal_ordering(
    terms: OperatorTermsList, weights: OperatorWeightsList = 1.0
) -> tuple[OperatorTermsList, OperatorWeightsList]:
    """
    Returns the normal ordered terms and weights of the fermion operator.
    """
    ordered_terms = []
    ordered_weights = []
    # loop over all the terms and weights and order each single term with corresponding weight
    for term, weight in zip(terms, weights):
        ordered = _normal_order_term(term, weight)
        ordered_terms += ordered[0]
        ordered_weights += ordered[1]
    ordered_terms = _make_tuple_tree(ordered_terms)
    return ordered_terms, ordered_weights


def _pair_order_term(
    term: OperatorTerm, weight: Number = 1.0
) -> tuple[OperatorTermsList, OperatorWeightsList]:
    """
    Return a pair ordered single term of the fermion operator.
    Pair ordering corresponds to placing first the highest indices on the right,
    and then making sure the creation operators are on the left and annihilation on the right.
    In this ordering, we make sure to account for the anti-commutation of operators.
    """

    parity = -1
    term = copy.deepcopy(list(term))
    weight = copy.copy(weight)
    if len(term) == 0:
        return [term], [weight]
    ordered_terms = []
    ordered_weights = []
    # the arguments given to this function will be transformed in a normal ordered way
    # loop through all the operators in the single term from left to right and order them
    # by swapping the term operators (and transform the weights by multiplying with the parity factors)
    for i in range(1, len(term)):
        for j in range(i, 0, -1):
            right_term = term[j]
            left_term = term[j - 1]

            ## exchange operators if biggest is on the right (need commutataion relations)
            # exchange operators if biggest is on the left (need commutataion relations)
            if right_term[0] < left_term[0]:  # not the same, always switch
                # fulfil anti commutation relations
                term[j - 1] = right_term
                term[j] = left_term
                weight *= parity

            # exchange operators if creation operator is on the right and annihilation on the left
            elif right_term[0] == left_term[0]:
                if right_term[1] and not left_term[1]:
                    term[j - 1] = right_term
                    term[j] = left_term
                    weight *= parity

                    # if same indices switch order (creation on the left), remember a a^ = 1 + a^ a
                    # if right_term[0] == left_term[0]:
                    new_term = term[: (j - 1)] + term[(j + 1) :]

                    # ad the processed term
                    o, w = _pair_order_term(tuple(new_term), parity * weight)
                    ordered_terms += o
                    ordered_weights += w

                # if we have two creation or two annihilation operators
                elif right_term[1] == left_term[1]:
                    # If same two Fermionic operators are repeated,
                    # evaluate to zero.
                    # if parity == -1 and right_term[0] == left_term[0]:
                    return ordered_terms, ordered_weights

    ordered_terms.append(term)
    ordered_weights.append(weight)
    return ordered_terms, ordered_weights


def _pair_ordering(
    terms: OperatorTermsList, weights: OperatorWeightsList = 1.0
) -> tuple[OperatorTermsList, OperatorWeightsList]:
    """
    Returns the pair ordered terms and weights of the fermion operator.
    """
    ordered_terms = []
    ordered_weights = []
    # loop over all the terms and weights and order each single term with corresponding weight
    for term, weight in zip(terms, weights):
        ordered = _pair_order_term(term, weight)
        ordered_terms += ordered[0]
        ordered_weights += ordered[1]
    ordered_terms = _make_tuple_tree(ordered_terms)
    return ordered_terms, ordered_weights


def _check_hermitian(
    terms: OperatorTermsList, weights: OperatorWeightsList, cutoff: float
) -> bool:
    """
    Check whether a set of terms and weights for a hermitian operator
    The terms are ordered into a canonical form with daggers and high orbital numbers to the left.
    After conjugation, the result is again reordered into canonical form.
    The result of both ordered lists of terms and weights are compared to be the same
    """
    # save in a dictionary the normal ordered terms and weights
    normal_ordered = _normal_ordering(terms, weights)
    dict_normal = defaultdict(complex)
    for term, weight in zip(*normal_ordered):
        dict_normal[tuple(term)] += weight

    # take the hermitian conjugate of the terms
    hc = _herm_conj(terms, weights)
    # normal order the h.c. terms
    hc_normal_ordered = _normal_ordering(*hc)

    # save in a dictionary the normal ordered h.c. terms and weights
    dict_hc_normal = defaultdict(complex)
    for term_hc, weight_hc in zip(*hc_normal_ordered):
        dict_hc_normal[tuple(term_hc)] += weight_hc

    # check if hermitian by comparing the dictionaries
    dict_normal = dict(dict_normal)
    dict_hc_normal = dict(dict_hc_normal)

    # compare dict up to a tolerance (1e-10)
    is_hermitian = _dict_compare(dict_normal, dict_hc_normal, cutoff)

    return is_hermitian


def _herm_conj(
    terms: OperatorTermsList, weights: OperatorWeightsList = 1
) -> tuple[OperatorTermsList, OperatorWeightsList]:
    """Returns the hermitian conjugate of the terms and weights."""
    conj_term = transpose_terms(terms)
    conj_weight = [np.conjugate(weight) for weight in weights]
    return conj_term, conj_weight


def transpose_terms(terms: OperatorTermsList) -> OperatorTermsList:
    """Returns the transpose of the terms (creation/destruction operators).
    The transpose is equivalent to the hermitian transpose because they are
    real.
    """
    conj_term = []
    for term in terms:
        conj_term.append([(op, 1 - int(dag)) for (op, dag) in reversed(term)])
    return conj_term


def transpose_term(term: OperatorTerm) -> OperatorTerm:
    """Returns the transpose of a single term (creation/destruction operators).
    The transpose is equivalent to the hermitian transpose because they are
    real.
    """
    # equivalent to transpose_terms
    return tuple((op, 1 - int(dag)) for (op, dag) in reversed(term))


def _convert_terms_to_spin_blocks(
    terms: OperatorTermsList, n_orbitals: int, n_spin_components: int
) -> OperatorTermsList:
    """
    See explanation in from_openfermion in conversion between conventions of netket
    and openfermion.

    Args:
        terms: the operator terms in tuple tree format
        n_orbitals: number of orbitals
        n_spin_components: number of spin components (2*spin+1)

    Returns:
        new terms tree
    """

    if n_spin_components == 1:
        return terms

    def _convert_loc(of_loc):
        orb_idx = of_loc // n_spin_components
        spin_idx = of_loc % n_spin_components
        return orb_idx + n_orbitals * spin_idx

    def _convert_term(term):
        return tuple([(_convert_loc(t[0]), t[1]) for t in term])

    return tuple(list(map(_convert_term, terms)))


def _canonicalize_input(
    terms: OperatorTermsList,
    weights: OperatorWeightsList,
    dtype: DType | None,
    cutoff: float,
    constant: Number = 0,
) -> tuple[OperatorDict, Number, DType]:
    r"""
    The canonical form is a tree tuple with a tuple pair of integers at the
    lowest level

    A term of the form :math:`\hat{a}_1^\dagger \hat{a}_2` would take the form
        `((1,1), (2,0))`, where (1,1) represents :math:`\hat{a}_1^\dagger` and (2,0)
        represents :math:`\hat{a}_2`.
    """
    if terms is None:
        terms = []

    if isinstance(terms, str):
        terms = (terms,)

    if len(terms) > 0:
        terms = _parse_term_tree(terms)

    if weights is None:
        weights = [1.0] * len(terms)

    weights = list(weights)
    # convert a constant to a diagonal operator
    if not np.isclose(constant, 0.0, atol=cutoff):
        terms = [()] + list(terms)
        weights = [constant] + weights

    dtype = canonicalize_dtypes(float, *weights, constant, dtype=dtype)

    weights = np.array(weights, dtype=dtype).tolist()

    if not len(weights) == len(terms):
        raise ValueError(
            f"length of weights should be equal, but received {len(weights)} and {len(terms)}"
        )

    _check_tree_structure(terms)

    # add the weights of terms that occur multiple times
    operators = zero_defaultdict(dtype)
    for t, w in zip(terms, weights):
        operators[t] += w
    operators = _remove_dict_zeros(dict(operators), cutoff)

    return operators, dtype


def _verify_input(hilbert, operators, raise_error=True) -> bool:
    """Check whether all input is valid"""
    terms = list(operators.keys())

    def _check_op(fop):
        v1 = 0 <= fop[0] < hilbert.size
        if not v1:
            if raise_error:
                raise ValueError(
                    f"Found invalid orbital index {fop[0]} for hilbert space {hilbert} of size {hilbert.size}"
                )
            return False
        v2 = fop[1] in (0, 1)
        if not v2:
            if raise_error:
                raise ValueError(
                    f"Found invalid character {fop[1]} for dagger, which should be 0 (no dagger) or 1 (dagger)."
                )
            return False
        return True

    def _check_term(term):
        return all(_check_op(t) for t in term)

    return all(_check_term(term) for term in terms)


def _remove_dict_zeros(d: dict, cutoff: float) -> dict:
    """Remove redundant zero values from a dictionary"""
    return {k: v for k, v in d.items() if np.abs(v) > cutoff}


def _parse_term_tree(terms: OperatorTermsList) -> OperatorTermsList:
    """Convert the terms tree into a canonical form of tuple tree of depth 3"""

    def _parse_branch(t):
        if isinstance(t, str):
            return _parse_string(t)
        elif hasattr(t, "__len__"):
            return tuple([_parse_branch(b) for b in t])
        else:
            return int(t)

    return _parse_branch(terms)


def _parse_string(s: str) -> OperatorTerm:
    """Parse strings such as '1^ 2' into a term form ((1, 1), (2, 0))"""
    s = s.strip()
    if s == "":
        return ()
    s = re.sub(" +", " ", s)
    terms = s.split(" ")
    processed_terms = []
    for term in terms:
        if term[-1] == "^":
            dagger = True
            term = term[:-1]
        else:
            dagger = False
        orb_nr = int(term)
        processed_terms.append((orb_nr, int(dagger)))
    return tuple(processed_terms)


def _dict_compare(d1: dict, d2: dict, cutoff: float) -> bool:
    """
    Compare two dicts and return True if their keys and values
    are all the same (up to some tolerance)
    """
    d1 = _remove_dict_zeros(d1, cutoff)
    d2 = _remove_dict_zeros(d2, cutoff)
    d1_keys = set(d1.keys())
    d2_keys = set(d2.keys())
    if d1_keys != d2_keys:
        return False
    # We checked that d1 and d2 have the same keys. Now check the values.
    return all(np.isclose(d1[o], d2[o], atol=cutoff) for o in d1_keys)


def _make_tuple_tree(terms: PyTree) -> PyTree:
    """Make tuples, so terms are hashable.

    Input could be e.g. a pytree of lists of lists,
    which we convert to tuples of tuples.
    """

    def _make_tuple(branch):
        if hasattr(branch, "__len__"):
            return tuple([_make_tuple(t) for t in branch])
        else:
            return int(branch)

    return _make_tuple(terms)


def _check_tree_structure(terms: OperatorTermsList) -> OperatorTermsList:
    """
    Check whether the terms structure is depth 3 everywhere
    and contains pairs of (idx, dagger) everywhere
    """

    def _descend(tree, current_depth, depths, pairs):
        if current_depth == 2 and hasattr(tree, "__len__"):
            pairs.append(len(tree) == 2)
        if hasattr(tree, "__len__"):
            for branch in tree:
                _descend(branch, current_depth + 1, depths, pairs)
        else:
            depths.append(current_depth)

    depths = []
    pairs = []
    _descend(terms, 0, depths, pairs)
    if not np.all(np.array(depths) == 3):
        raise ValueError(f"terms is not a depth 3 tree, found depths {depths}")
    if not np.all(pairs):
        raise ValueError(
            "terms should be provided in (i, dag) pairs or empty for a constant"
        )


def _is_diag_term(term: OperatorTerm) -> bool:
    """
    Check whether this term changes the sample or not
    """

    def _init_empty_arr():
        return [
            0,
            0,
        ]  # first one counts number of daggers, second number of non-daggers

    if len(term) == 0:
        return True  # constant

    ops = defaultdict(_init_empty_arr)
    for orb_idx, dagger in term:
        ops[orb_idx][int(dagger)] += 1
    return all((x[0] == x[1]) for x in ops.values())


def zero_defaultdict(dtype: DType) -> defaultdict:
    """
    Temporary function to make sure we get a good initializer for each dtype
    """

    def _dtype_init():
        return np.array(0, dtype=dtype).item()

    return defaultdict(_dtype_init)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING
from functools import wraps

import numpy as np
from numba import jit


from netket.hilbert import AbstractHilbert, HomogeneousHilbert
from netket.errors import concrete_or_error, NumbaOperatorGetConnDuringTracingError
from netket.utils.types import DType

from .base import PauliStringsBase
from .jax import pack_internals

if TYPE_CHECKING:
    from .jax import PauliStringsJax


def pack_internals_numba(
    hilbert: AbstractHilbert,
    operators: dict,
    weights,
    dtype: DType,
    cutoff: float,  # unused
):
    acting = pack_internals(operators, weights)

    # the most Z we need to do anywhere
    _n_z_check_max = 0
    for v in acting.values():
        for _, b_z_check in v:
            _n_z_check_max = max(_n_z_check_max, len(b_z_check))

    n_operators = len(acting)
    # maximum number of strings which have the same sites to act on with X, but have different sites for Z
    _n_op_max = max(
        list(map(lambda x: len(x), list(acting.values()))), default=n_operators
    )

    # unpacking the dictionary into fixed-size arrays

    # the sites each X string is acting on, padded
    _sites = np.empty((n_operators, hilbert.size), dtype=np.intp)
    # number of sites each X string is acting on
    _ns = np.empty((n_operators), dtype=np.intp)
    # the number of operators for the same X string, with different sites we need to apply Z on
    _n_op = np.empty(n_operators, dtype=np.intp)
    # weights, padded
    _weights = np.empty((n_operators, _n_op_max), dtype=dtype)
    # the number of Z sites in each operators of the X strings, padded
    _nz_check = np.empty((n_operators, _n_op_max), dtype=np.intp)
    # sites to act on with Z for each operators of the X strings, padded
    _z_check = np.empty((n_operators, _n_op_max, _n_z_check_max), dtype=np.intp)

    for i, act in enumerate(acting.items()):
        sites = act[0]
        nsi = len(sites)
        _sites[i, :nsi] = sites
        _ns[i] = nsi
        values = act[1]
        _n_op[i] = len(values)
        for j in range(_n_op[i]):
            _weights[i, j] = values[j][0]
            _nz_check[i, j] = len(values[j][1])
            _z_check[i, j, : _nz_check[i, j]] = values[j][1]

    return {
        "sites": _sites,
        "ns": _ns,
        "n_op": _n_op,
        "weights_numba": _weights,
        "nz_check": _nz_check,
        "z_check": _z_check,
        "n_operators": n_operators,
        "mel_dtype": dtype,
    }


class PauliStrings(PauliStringsBase):
    """A Hamiltonian consisting of the sum of products of Pauli operators."""

    @wraps(PauliStringsBase.__init__)
    def __init__(
        self,
        hilbert: AbstractHilbert,
        operators: None | str | list[str] = None,
        weights: None | float | complex | list[float | complex] = None,
        *,
        cutoff: float = 1.0e-10,
        dtype: DType | None = None,
    ):
        super().__init__(hilbert, operators, weights, cutoff=cutoff, dtype=dtype)

        # check that it is homogeneous, throw error if it's not
        if not isinstance(self.hilbert, HomogeneousHilbert):
            local_states = self.hilbert.states_at_index(0)
            if not all(
                np.allclose(local_states, self.hilbert.states_at_index(i))
                for i in range(self.hilbert.size)
            ):
                raise ValueError(
                    "Hilbert spaces with non homogeneous local_states are not "
                    "yet supported by PauliStrings."
                )

        self._initialized = False

    def to_jax_operator(self) -> "PauliStringsJax":  # noqa: F821
        """
        Returns the jax-compatible version of this operator, which is an
        instance of :class:`netket.operator.PauliStringsJax`.
        """
        from .jax import PauliStringsJax

        return PauliStringsJax(
            self.hilbert,
            self.operators,
            self.weights,
            dtype=self.dtype,
            cutoff=self._cutoff,
        )

    @property
    def max_conn_size(self) -> int:
        """The maximum number of non zero ⟨x|O|x'⟩ for every x."""
        # 1 connection for every operator X, Y, Z...
        self._setup()
        return self._n_operators

    def _setup(self, force=False):
        """Analyze the operator strings and precompute arrays for get_conn inference"""
        if force or not self._initialized:
            data = pack_internals_numba(
                self.hilbert, self.operators, self.weights, self.dtype, self._cutoff
            )

            self._sites = data["sites"]
            self._ns = data["ns"]
            self._n_op = data["n_op"]
            self._weights_numba = data["weights_numba"]
            self._nz_check = data["nz_check"]
            self._z_check = data["z_check"]
            self._n_operators = data["n_operators"]

            # caches for execution
            self._x_prime_max = np.empty((self._n_operators, self.hilbert.size))
            self._mels_max = np.empty((self._n_operators), dtype=data["mel_dtype"])

            self._initialized = True

    def _reset_caches(self):
        super()._reset_caches()
        self._initialized = False

    @staticmethod
    @jit(nopython=True)
    def _flattened_kernel(
        x,
        sections,
        x_prime,
        mels,
        sites,
        ns,
        n_op,
        weights,
        nz_check,
        z_check,
        cutoff,
        max_conn,
        pad=False,
    ):
        x_prime = np.empty((x.shape[0] * max_conn, x_prime.shape[1]), dtype=x.dtype)
        mels = np.zeros((x.shape[0] * max_conn), dtype=mels.dtype)
        state_1 = 1

        n_c = 0
        for b in range(x.shape[0]):
            xb = x[b]
            # initialize with the old state
            x_prime[b * max_conn : (b + 1) * max_conn, :] = np.copy(xb)

            for i in range(sites.shape[0]):  # iterate over the X strings
                mel = 0.0
                # iterate over the Z substrings
                for j in range(n_op[i]):
                    # apply all the Z (check the qubits at all affected sites)
                    if nz_check[i, j] > 0:
                        to_check = z_check[i, j, : nz_check[i, j]]
                        n_z = np.count_nonzero(xb[to_check] == state_1)
                    else:
                        n_z = 0
                    # multiply with -1 for every site we did Z which was state_1
                    mel += weights[i, j] * (-1.0) ** n_z

                if cutoff is None or abs(mel) > cutoff:
                    x_prime[n_c] = np.copy(xb)
                    # now flip all the sites in the X string
                    for site in sites[i, : ns[i]]:
                        new_state_idx = int(x_prime[n_c, site] == 0)
                        x_prime[n_c, site] = new_state_idx
                    mels[n_c] = mel
                    n_c += 1

            if pad:
                n_c = (b + 1) * max_conn

            sections[b] = n_c
        return x_prime[:n_c], mels[:n_c]

    def get_conn_flattened(self, x, sections, pad=False):
        r"""Finds the connected elements of the Operator. Starting
        from a given quantum number x, it finds all other quantum numbers x' such
        that the matrix element :math:`O(x,x')` is different from zero. In general there
        will be several different connected states x' satisfying this
        condition, and they are denoted here :math:`x'(k)`, for :math:`k=0,1...N_{\mathrm{connected}}`.

        This is a batched version, where x is a matrix of shape (batch_size,hilbert.size).

        Args:
            x (matrix): A matrix of shape (batch_size,hilbert.size) containing
                        the batch of quantum numbers x.
            sections (array): An array of size (batch_size) useful to unflatten
                        the output of this function.
                        See numpy.split for the meaning of sections.

        Returns:
            matrix: The connected states x', flattened together in a single matrix.
            array: An array containing the matrix elements :math:`O(x,x')` associated to each x'.

        """
        self._setup()

        x_ids = self.hilbert.states_to_local_indices(x)
        x_ids = concrete_or_error(
            np.asarray,
            x_ids,
            NumbaOperatorGetConnDuringTracingError,
            self,
        )

        assert (
            x.shape[-1] == self.hilbert.size
        ), "size of hilbert space does not match size of x"
        xp_ids, mels = self._flattened_kernel(
            x_ids,
            sections,
            self._x_prime_max,
            self._mels_max,
            self._sites,
            self._ns,
            self._n_op,
            self._weights_numba,
            self._nz_check,
            self._z_check,
            self._cutoff,
            self._n_operators,
            pad,
        )
        xp = self.hilbert.local_indices_to_states(xp_ids, dtype=x.dtype)
        return xp, mels



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Callable
from functools import partial

import jax
import jax.numpy as jnp

from netket import jax as nkjax
from netket.utils.types import Array, PyTree


@partial(jax.jit, static_argnames=("apply_fun"))
def is_probably_holomorphic(
    apply_fun: Callable[[PyTree, Array], Array],
    parameters: PyTree,
    samples: Array,
    model_state: PyTree | None = None,
) -> bool:
    r"""
    Check if a function :math:`\psi` is likely to be holomorphic almost
    everywhere.

    The check is done by verifying if the function satisfies on the provided
    samples the
    `Cauchy-Riemann equations <https://en.wikipedia.org/wiki/Cauchy–Riemann_equations>`_.
    In particular, assuming that the function is complex-valued
    :math:`\psi(\theta, x) = \psi_r(\theta, x) + i \psi_i(\theta, x)`
    where :math:`\psi_r` and :math:`\psi_i` are it's real and imaginary part,
    and the parameters can also be split in real and imaginary part according to
    :math:`\theta = \theta_r + i\theta_i`, the conditions real

    .. math::

        \frac{\partial \psi_r(\theta, x)}{\partial \theta_r} =
        \frac{\partial \psi_i(\theta, x)}{\partial \theta_i}
        \,\,\,\,\,\,\,\,\,\,\,\,\,\,\,\,\,\,\,\,\,\,\,\,\,\,\,\,
        \frac{\partial \psi_r(\theta, x)}{\partial \theta_i} =
        -\frac{\partial \psi_i(\theta, x)}{\partial \theta_r}

    Those conditions are always not verified if the parameters are
    real (:math:`\theta_i=0`), in which case we automatically return False.

    We verify the Cauchy-Riemann equations on the provided samples, so you should
    provide a sufficiently large number of points to be relatively sure if your
    function is holomorphic, because there are some functions that are locally non
    holomorphic, however that is rare and in general you can trust the result of this
    function.

    .. Note::

        If you are working with NetKet, you are most likely pretty good at math. In that
        case, instead of relying on the computer to tell you if the function is holomorphic
        or not you can do it by yourself! It's quite easy: the Cauchy-Riemann conditions
        can be rewritten using `Wirtinger derivatives <https://en.wikipedia.org/wiki/Wirtinger_derivatives>`_
        (where we treat the complex conjugate variables :math:`\theta` and :math:`\theta^\star`
        as linearly independent) as

        .. math::

            \frac{\partial \psi(\theta, x)]}{\partial \theta^\star} = 0 .

        Practically, this can be read as *if the function :math:`\psi` depends on
        :math:`\theta^\star` it will not be holomorphic.* Therefore, a very
        easy way to spot non-holomorphic functions is to verify if any operation
        among the following appear in the function:

        * complex conjugation
        * absolute value
        * real or imaginary part


    .. Note::

        In most Machine-Learning applications we are only interested in holomorphicity
        almost-everywhere, therefore if the function is not holomorphic on one particular
        point, such as :math:`\theta=0`, :math:`\theta=1` or similar this will in general
        not matter because that is a set of dimension zero that we will never cross.

        This is similar to the way the derivative of locally non-differentiable functions like
        the modulus is obtained: we assume that we never enter those regions.

    Args:
        apply_fun: The forward pass of the variational function. This should be a Callable
            accepting two inputs, the first being the parameters as a dictionary with at least
            a key :code:`params`, and possibly other keys given by the :code:`model_state`. The
            second argument to :code:`apply_fun` will be the samples.
        parameters : a pytree of parameters to be passed as the first argument to the :code:`apply_fun`.
            We will check if the ansatz is holomoprhic with respect to the derivatives of those
            parameters.
        samples : a set of samples.
        model_state: optional dictionary of state parameters of the model.

    """

    if nkjax.tree_leaf_isreal(parameters):
        return False

    samples = samples.reshape(-1, samples.shape[-1])
    jacs = nkjax.jacobian(
        apply_fun, parameters, samples, model_state=model_state, mode="complex"
    )

    # ∂ᵣψᵣ
    dr_dpr = jax.tree_util.tree_map(lambda x: x[:, 0, ...], jacs.real)
    # ∂ᵣψᵢ
    dr_dpi = jax.tree_util.tree_map(lambda x: x[:, 1, ...], jacs.real)
    # ∂ᵢψᵣ
    di_dpr = jax.tree_util.tree_map(lambda x: x[:, 0, ...], jacs.imag)
    # ∂ᵢψᵢ
    di_dpi = jax.tree_util.tree_map(lambda x: x[:, 1, ...], jacs.imag)

    # verify that ∂ᵣψᵣ == ∂ᵢψᵢ
    cond1 = jax.tree_util.tree_map(lambda x, y: jnp.allclose(x, y), dr_dpr, di_dpi)
    # verify that ∂ᵣψᵢ == -∂ᵢψᵣ
    cond2 = jax.tree_util.tree_map(lambda x, y: jnp.allclose(x, -y), dr_dpi, di_dpr)

    return jax.tree_util.tree_reduce(jnp.bitwise_and, (cond1, cond2))  # type: ignore



# Copyright 2021 The NetKet Authors - All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import IntFlag, auto
from typing import TYPE_CHECKING

import jax
import jax.numpy as jnp

from netket.utils import struct, KahanSum
from netket.utils.types import Array

if TYPE_CHECKING:
    from ._solver import SolverState


class IntegratorFlags(IntFlag):
    """
    Enum class containing flags for signaling solver information from within `jax.jit`ed code.
    """

    NONE = 0
    INFO_STEP_ACCEPTED = auto()
    WARN_MIN_DT = auto()
    WARN_MAX_DT = auto()
    ERROR_INVALID_DT = auto()

    WARNINGS_FLAGS = WARN_MIN_DT | WARN_MAX_DT
    ERROR_FLAGS = ERROR_INVALID_DT

    __MESSAGES__ = {
        INFO_STEP_ACCEPTED: "Step accepted",
        WARN_MIN_DT: "dt reached lower bound",
        WARN_MAX_DT: "dt reached upper bound",
        ERROR_INVALID_DT: "Invalid value of dt",
    }

    def message(self) -> str:
        """Returns a string with a description of the currently set flags."""
        msg = self.__MESSAGES__
        return ", ".join(msg[flag] for flag in msg.keys() if flag & self != 0)


class IntegratorState(struct.Pytree):
    r"""
    Dataclass containing the state of an ODE solver.
    In particular, it stores the current state of the system, former usefull values
    and information about integration (number of step, errors, etc)
    """

    solver_state: "SolverState"
    """The state of the solver."""

    step_no: int
    """The number of successful steps since the start of the iteration."""
    step_no_total: int
    """The number of steps since the start of the iteration, including rejected steps."""
    t: KahanSum
    """The current time."""
    y: Array
    """The solution at current time."""
    dt: float
    """The current time-step size."""
    last_norm: float | None = None
    """The solution norm at previous time step."""
    last_scaled_error: float | None = None
    """The error of the TDVP integrator at the last time step."""
    flags: IntegratorFlags = IntegratorFlags.INFO_STEP_ACCEPTED
    """The flags containing information on the solver state."""

    def __init__(
        self,
        dt: float,
        y,
        t,
        solver,
        *,
        step_no=0,
        step_no_total=0,
        last_norm=None,
        last_scaled_error=None,
        flags=IntegratorFlags.INFO_STEP_ACCEPTED,
    ):
        r"""
        Args:
            dt: The current time-step size.
            y: The solution at current time.
            t: The current time.
            solver: The ODE solver.

            step_no: The number of successful steps since the start of the iteration.
                defaults to :code:`0`.
            step_no_total: The number of steps since the start of the iteration, including rejected steps.
                defaults to :code:`0`.
            last_norm: The solution norm at previous time step.
            last_scaled_error: The error of the TDVP integrator at the last time step.
            flags: The flags containing information on the solver state.
        """
        step_dtype = jnp.int64 if jax.config.x64_enabled else jnp.int32
        err_dtype = jnp.float64 if jax.config.x64_enabled else jnp.float32

        self.step_no = jnp.asarray(step_no, dtype=step_dtype)
        self.step_no_total = jnp.asarray(step_no_total, dtype=step_dtype)

        if not isinstance(t, KahanSum):
            t = KahanSum(t)
        self.t = t
        if not isinstance(dt, jax.Array):
            dt = jnp.asarray(dt)
        self.dt = dt

        # To avoid creating problems, just convert to array if the leaves
        # in the parameters are not jax arrays already.
        def _maybe_asarray(x):
            if isinstance(x, jax.Array):
                return x
            else:
                return jnp.asarray(x)

        self.y = jax.tree_util.tree_map(_maybe_asarray, y)

        if last_norm is not None:
            last_norm = jnp.asarray(last_norm, dtype=err_dtype)
        self.last_norm = last_norm
        if last_scaled_error is not None:
            last_scaled_error = jnp.asarray(last_scaled_error, dtype=err_dtype)
        self.last_scaled_error = last_scaled_error

        # Todo: if making SolverFlag a proper pytree this can be restored.
        # if not isinstance(flags, SolverFlags):
        #    raise TypeError(f"flags must be SolverFlags but got {type(flags)} : {flag}")
        self.flags = flags

        self.solver_state = solver._init_state(self)

    def __repr__(self) -> str:
        try:
            dt = f"{self.dt:.2e}"
            last_norm = f", {self.last_norm:.2e}" if self.last_norm is not None else ""
            accepted = f", {'A' if self.accepted else 'R'}"
        except (ValueError, TypeError):
            dt = f"{self.dt}"
            last_norm = f"{self.last_norm}"
            accepted = f"{IntegratorFlags.INFO_STEP_ACCEPTED}"

        solver_state = f", solver_state={self.solver_state}"

        return f"IntegratorState(step_no(total)={self.step_no}({self.step_no_total}), t={self.t.value}, dt={dt}{last_norm}{accepted}{solver_state})"

    @property
    def accepted(self) -> bool:
        """Boolean indicating whether the last step was accepted."""
        return IntegratorFlags.INFO_STEP_ACCEPTED & self.flags != 0



# this file contains the logic of the operators, essentially _jw_kernel

from typing import Union

from functools import partial

import numpy as np
import itertools

import jax
import jax.numpy as jnp

from netket.jax import COOArray
from netket.utils.types import Array

from ._operator_data import PNCOperatorDataType


def _comb(kl: Array, n: int) -> Array:
    r"""
    compute all combinations of n elements from a set kl
    Args:
        kl: 1d array of elements
        n: size n
    Returns:
        an array containg all combinations of size n of elemenets in kl
    """
    if len(kl) < n:
        return jnp.zeros((n, 0), dtype=kl.dtype)
    c = list(itertools.combinations(np.arange(len(kl)), n))
    return kl[np.array(c, dtype=kl.dtype).T[::-1]]


def _jw_kernel(
    k_destroy: Array, l_create: Array, x: Array
) -> tuple[Array, Array, Array]:
    r"""
    compute all matrix elements :math:`x^\prime` such that :math:`\langle x^\prime | \hat c^\dagger_{l_m} \cdots \hat c^\dagger_{l_1} \hat c_{k_n} \cdots \hat c_{k_1} | x\rangle \neq 0`
    of a batch of states x.

    Args:
        k_destroy: an array of indices of the :math:`\hat c`
        l_create: an array of indices of the :math:`\hat c^\dagger`
        x: an array of states
    Returns:
        xp: matrix elements :math:`x^\prime`
        sign: sign (value) of the matrix element if it is nonzero, arbitrary value otherwise
        create_was_empty: if the matrix element is zero
    """
    # destroy
    xd = jax.vmap(lambda i: x.at[i].set(0))(k_destroy.T)
    # create
    xp = jax.vmap(jax.vmap(lambda x, i: x.at[i].set(1), in_axes=(None, 0)))(
        xd, l_create
    )

    m = jnp.arange(x.shape[-1], dtype=k_destroy.dtype)

    # we apply the destruction operators in descending order,
    # the jordan-wigner sign of an operator does not depend on sites larger than it, therefore,
    # given it is in normal order, we can compute it all in terms of the initial state.
    # (sum the axis which is the one of the indices we destroy/create (size number of operators//2))
    jw_mask_destroy = jax.lax.reduce_xor(k_destroy[..., None] > m, axes=(0,))

    # same for when we create again, except then have to apply it to the state where we already destroyed
    jw_mask_create = jax.lax.reduce_xor(l_create[..., None] > m, axes=(2,))

    create_was_empty = jax.vmap(jax.vmap(lambda x, i: ~x[i].any(), in_axes=(None, 0)))(
        xd, l_create
    )

    sgn_destroy = jax.lax.reduce_xor(
        jw_mask_destroy * x[None], axes=((jw_mask_destroy * x[None]).ndim - 1,)
    )
    sgn_create = jax.lax.reduce_xor(
        jw_mask_create * xd[:, None], axes=((jw_mask_create * xd[:, None]).ndim - 1,)
    )
    sgn = sgn_create + sgn_destroy[:, None]
    sgn = jax.lax.bitwise_and(sgn, jnp.ones_like(sgn)).astype(bool)
    sign = 1 - 2 * sgn.astype(np.int8)

    return xp, sign, create_was_empty


@partial(jax.jit, static_argnums=0)
@partial(jnp.vectorize, signature="(n)->(m,n),(m)", excluded=(0, 2, 3, 4))
def _get_conn_padded(
    n_fermions: int,
    x: Array,
    index_array: Union[Array, COOArray, None],
    create_array: Union[Array, None],
    weight_array: Array,
) -> tuple[Array, Array]:
    r"""
    helper function for the matrix elements functions defined below

    does not know about spin sectors

    Args:
        n_fermionsnumber of electrons
        x: occupation vectors
        index_array, create_array, weight_array: internal (sparse) operator data representation
    Returns:
        connected states and corresponding matrix elements
    """
    assert x.ndim == 1
    if index_array is not None:
        half_n_ops = index_array.ndim
    else:  # diagonal
        half_n_ops = weight_array.ndim

    if half_n_ops == 0:  # constant
        xp = x[None, :]
        mels = weight_array.reshape(xp.shape[:-1])
    else:
        dtype = x.dtype

        (l_occupied,) = jnp.where(x, size=n_fermions)
        k_destroy = _comb(l_occupied, half_n_ops)

        if index_array is None:  # diagonal
            weight = weight_array[tuple(k_destroy)]
            xp = x[None, :]
            # we first destroy in desc order, then create
            # sites not acted on cancel by the create/destroy pair of the same site,
            # so we can assume they are not there.
            # When we create all smaller sites acted on are 0,
            # therefore the jw sign is determined just from the signs from destroy.
            # Then it' is easy to see that only every other site counts (the rest cancel),
            # and the sign is given by (+1 if there is an even number of other sites, -1 if odd)
            # sign = [+,+,-,-,+,+,-,-,+,+,-,-,...][half_n_ops]
            sgn = (half_n_ops // 2) % 2
            sign = 1 - 2 * sgn
            mels = sign * weight.sum()[None]
        else:
            ind = index_array[tuple(k_destroy)]
            weight = weight_array[ind]
            l_create = create_array[ind]

            xp, sign, create_was_empty = _jw_kernel(k_destroy, l_create, x)
            mels = weight * sign * create_was_empty

            # make sure we don't return states w/ wrong number of electrons
            # because of the padding we check if the mel is 0
            # xp = jnp.where(create_was_empty[..., None], xp, x[..., None, None, :])
            xp = jnp.where((mels == 0)[:, :, None], x[None, None, :], xp)

            xp = jax.lax.collapse(xp, 0, xp.ndim - 1).astype(dtype)
            mels = jax.lax.collapse(mels, 0, mels.ndim)
    return xp, mels


@partial(jax.jit, static_argnames="n_spin_subsectors")
def unpack_spin_sectors(x: Array, n_spin_subsectors: int = 2):
    r"""
    split spin sectors of x

    Args:
        a single stacked array of occupations
    Returns:
        one array for each spin sector
    """
    assert x.shape[-1] % n_spin_subsectors == 0
    x_ = x.reshape(x.shape[:-1] + (n_spin_subsectors, x.shape[-1] // n_spin_subsectors))
    return tuple(x_[..., i, :] for i in range(n_spin_subsectors))


@jax.jit
def pack_spin_sectors(*xs: tuple[Array]) -> Array:
    r"""
    flatten spin sectors of xs

    Args:
        xs: one array of occupations for each spin sector
    Returns:
        a single stacked array
    """
    xs = jnp.broadcast_arrays(*xs)
    xd = xs[0]
    n_spin_subsectors = len(xs)
    res = jnp.zeros(
        xd.shape[:-1]
        + (
            n_spin_subsectors,
            xd.shape[-1],
        ),
        dtype=xd.dtype,
    )
    for i, xi in enumerate(xs):
        res = res.at[..., i, :].set(xi)
    return jax.lax.collapse(res, res.ndim - 2, res.ndim)


@partial(jax.jit, static_argnums=(0, 1))
@partial(jnp.vectorize, signature="(n),(n)->(m,n),(m,n),(m)", excluded=(0, 1, 4, 5, 6))
def _get_conn_padded_interaction_up_down(
    nelectron_down: int,
    nelectron_up: int,
    x_down: Array,
    x_up: Array,
    index_array: Union[Array, COOArray, None],
    create_array: Union[Array, None],
    weight_array: Array,
) -> tuple[Array, Array, Array]:
    r"""
    helper function for the matrix elements for a 2-body interaction term in two different spin sectors
    i.e. of the operator :math:`\sum_{ijkl} w_{ijkl} \hat c^\dagger_{i\downarrow}  \hat c^\dagger_{j_downarrow} \hat c^\dagger_{k_uparrow} \hat c^\dagger_{l_uparrow}`

    Args:
        nelectron_down, nelectron_up: number of electrons in the down and up sector
        x_down, x_up: occupation vectors in both sectors
        index_array, create_array, weight_array: internal (sparse) operator data representation
    Returns:
        connected states and corresponding matrix elements
    """
    dtype = x_down.dtype

    assert x_down.ndim == 1
    if index_array is not None:
        assert index_array.ndim == 2
    else:  # diagonal
        assert weight_array.ndim == 2

    (down_occupied,) = jnp.where(x_down, size=nelectron_down)
    (up_occupied,) = jnp.where(x_up, size=nelectron_up)

    k_destroy_down, k_destroy_up = jnp.meshgrid(down_occupied, up_occupied)

    if index_array is None:  # diagonal
        weight = weight_array[k_destroy_down, k_destroy_up]
        xp_down = x_down[None, :]
        xp_up = x_up[None, :]
        sign = 1
        mels = sign * weight.sum()[None]
    else:
        ind = index_array[k_destroy_down, k_destroy_up].ravel()
        weight = weight_array[ind]
        l_create = create_array[ind]

        k_destroy_down = k_destroy_down.reshape(1, -1)
        k_destroy_up = k_destroy_up.reshape(1, -1)
        l_create_down = l_create[..., :1]
        l_create_up = l_create[..., 1:]

        xp_down, sign_down, down_create_is_not_occupied = _jw_kernel(
            k_destroy_down, l_create_down, x_down
        )
        xp_up, sign_up, up_create_is_not_occupied = _jw_kernel(
            k_destroy_up, l_create_up, x_up
        )

        up_is_diagonal = k_destroy_up[0][:, None] == l_create_up[..., 0]
        down_is_diagonal = k_destroy_down[0][:, None] == l_create_down[..., 0]
        both_not_occupied = (down_create_is_not_occupied | down_is_diagonal) & (
            up_create_is_not_occupied | up_is_diagonal
        )

        sign = sign_up * sign_down
        mels = weight * both_not_occupied * sign

        xp_down = jnp.where((mels == 0)[:, :, None], x_down[None, None, :], xp_down)
        xp_up = jnp.where((mels == 0)[:, :, None], x_up[None, None, :], xp_up)

        xp_down = jax.lax.collapse(xp_down, 0, xp_down.ndim - 1).astype(dtype)
        xp_up = jax.lax.collapse(xp_up, 0, xp_up.ndim - 1).astype(dtype)
        mels = jax.lax.collapse(mels, 0, mels.ndim)
    return xp_down, xp_up, mels


@partial(jax.jit, static_argnames=("n_fermions",))
def get_conn_padded_pnc(
    _operator_data: PNCOperatorDataType, x: Array, n_fermions: int
) -> tuple[Array, Array]:
    r"""
    compute the connected elements for ParticleNumberConservingFermioperator2nd

    Args:
        _operator_data: internal sparse operator representation
        x: occupation vectors
        n_fermions: number of electrons
    Returns:
        connected states and corresponding matrix elements
    """
    dtype = x.dtype
    if not jnp.issubdtype(dtype, jnp.integer) or jnp.issubdtype(dtype, jnp.integer):
        x = x.astype(jnp.int8)

    xp_list = []
    mels_list = []
    xp_diag = None
    mels_diag = 0
    for k, v in _operator_data["diag"].items():
        xp, mels = _get_conn_padded(n_fermions, x, *v)
        xp_diag = xp
        mels_diag = mels_diag + mels
        xp_list = [xp_diag]
        mels_list = [mels_diag]
    for k, v in _operator_data["offdiag"].items():
        xp, mels = _get_conn_padded(n_fermions, x, *v)
        xp_list.append(xp)
        mels_list.append(mels)
    xp = jnp.concatenate(xp_list, axis=-2)
    mels = jnp.concatenate(mels_list, axis=-1)
    return xp.astype(dtype), mels


@partial(jax.jit, static_argnames=("n_fermions_per_spin",))
def get_conn_padded_pnc_spin(
    _operator_data: PNCOperatorDataType, x: Array, n_fermions_per_spin: tuple[int]
) -> tuple[Array, Array]:
    r"""
    compute the connected elements for ParticleNumberAndSpinConservingFermioperator2nd

    Args:
        _operator_data: internal sparse operator representation
        x: occupation vectors (with concatenated spin sectors)
        n_fermions_per_spin: number of electrons in each spin sector
    Returns:
        connected states and corresponding matrix elements
    """
    n_spin_subsectors = len(n_fermions_per_spin)
    xs = unpack_spin_sectors(x, n_spin_subsectors)
    xs_diag = tuple(a[..., None, :] for a in xs)
    dtype = xs[0].dtype

    xp_list = []
    mels_list = []
    xp_diag = None
    mels_diag = 0

    # TODO make sectors a jax array and use jax loop here to compile only once ?
    # (requires n_fermions_per_spin to be the same for all sectors)

    for (k, sectors), v in _operator_data["diag"].items():
        if k == 0:
            assert sectors == ()
            sectors = (0,)  # dummy sector

        for i in sectors:
            # act on a single sector: we use the non-spin mel code here for the mels
            _, melsi = _get_conn_padded(n_fermions_per_spin[i], xs[i], *v)
            mels_diag = mels_diag + melsi
            xp_diag = x[..., None, :]

    for (k, sectors), v in _operator_data["mixed_diag"].items():
        if k != 4:
            raise NotImplementedError
        # TODO make sectors a jax array and use jax loop here to compile only once
        for i, j in sectors:
            assert i > j  # here i>j
            # e.g. take operator data to be c_ijkl + c_jilk so that here we only need to sum  ρ > σ (i.e. σ=d, ρ=u)
            *_, melsij = _get_conn_padded_interaction_up_down(
                n_fermions_per_spin[j], n_fermions_per_spin[i], xs[j], xs[i], *v
            )
            mels_diag = mels_diag + melsij
            xp_diag = x[..., None, :]

    # TODO optionally always add zero diagonal?
    if xp_diag is not None:
        xp_list = [xp_diag]
        mels_list = [mels_diag]

    for (k, sectors), v in _operator_data["offdiag"].items():
        # TODO make sectors a jax array and use jax loop here to compile only once ?
        # (requires n_fermions_per_spin to be the same for all sectors)
        for i in sectors:
            # act on a single sector: we use the non-spin mel code here, and just apply an identity in the other sectors
            xpi, melsi = _get_conn_padded(n_fermions_per_spin[i], xs[i], *v)
            xpi = pack_spin_sectors(*xs_diag[:i], xpi, *xs_diag[i + 1 :])
            xp_list.append(xpi)
            mels_list.append(melsi)

    for (k, sectors), v in _operator_data["mixed_offdiag"].items():
        if k != 4:
            raise NotImplementedError
        for i, j in sectors:
            assert i > j  # here i>j
            # e.g. take operator data to be c_ijkl + c_jilk so that here we only need to sum  ρ > σ (i.e. σ=d, ρ=u)
            xpj, xpi, melsij = _get_conn_padded_interaction_up_down(
                n_fermions_per_spin[j], n_fermions_per_spin[i], xs[j], xs[i], *v
            )
            xpij = pack_spin_sectors(
                *xs_diag[:j], xpj, *xs_diag[j + 1 : i], xpi, *xs_diag[i + 1 :]
            )
            xp_list.append(xpij)
            mels_list.append(melsij)
    if len(xp_list) > 0:
        xp = jnp.concatenate(xp_list, axis=-2).astype(dtype)
        mels = jnp.concatenate(mels_list, axis=-1)
    else:
        xp = jnp.zeros((*x.shape[:-1], 0, x.shape[-1]), dtype=dtype)
        mels = jnp.zeros(xp.shape[:-1])  # TODO dtype?
    return xp, mels



from collections.abc import Callable

import jax
import jax.numpy as jnp

from ._chunk_utils import _chunk, _unchunk
from ._scanmap import scanmap, scan_append

from netket.utils import HashablePartial
from netket.utils import config
from netket.jax.sharding import sharding_decorator


def _eval_fun_in_chunks(vmapped_fun, chunk_size, argnums, *args, **kwargs):
    n_elements = jax.tree_util.tree_leaves(args[argnums[0]])[0].shape[0]
    n_chunks, n_rest = divmod(n_elements, chunk_size)

    if n_chunks == 0 or chunk_size >= n_elements:
        y = vmapped_fun(*args, **kwargs)
    else:
        # split inputs
        def _get_chunks(x):
            x_chunks = jax.tree_util.tree_map(
                lambda x_: x_[: n_elements - n_rest, ...], x
            )
            x_chunks = _chunk(x_chunks, chunk_size)
            return x_chunks

        def _get_rest(x):
            x_rest = jax.tree_util.tree_map(
                lambda x_: x_[n_elements - n_rest :, ...], x
            )
            return x_rest

        args_chunks = [
            _get_chunks(a) if i in argnums else a for i, a in enumerate(args)
        ]
        args_rest = [_get_rest(a) if i in argnums else a for i, a in enumerate(args)]

        y_chunks = _unchunk(
            scanmap(vmapped_fun, scan_append, argnums)(*args_chunks, **kwargs)
        )

        if n_rest == 0:
            y = y_chunks
        else:
            y_rest = vmapped_fun(*args_rest, **kwargs)
            y = jax.tree_util.tree_map(
                lambda y1, y2: jnp.concatenate((y1, y2)), y_chunks, y_rest
            )
    return y


def _eval_fun_in_chunks_sharding(vmapped_fun, chunk_size, argnums, *args, **kwargs):
    # Equivalent to `_eval_fun_in_chunks` above but preserves sharding,
    # by computing the vmapped_fun in chunks on every shard (which sits on a separate device)
    sharded_args_tree = tuple(i in argnums for i, a in enumerate(args))
    f = HashablePartial(_eval_fun_in_chunks, vmapped_fun, chunk_size, argnums, **kwargs)
    return sharding_decorator(f, sharded_args_tree)(*args)


def _chunk_vmapped_function(
    vmapped_fun: Callable,
    chunk_size: int | None,
    argnums: int | tuple[int, ...] = 0,
    axis_0_is_sharded: bool = False,
) -> Callable:
    """takes a vmapped function and computes it in chunks"""

    if chunk_size is None:
        # TODO: Here we are not applyign sharding_decorator, while below we are
        # Maybe we should always apply it for consistency?
        return vmapped_fun

    if isinstance(argnums, int):
        argnums = (argnums,)
    if axis_0_is_sharded:
        return HashablePartial(
            _eval_fun_in_chunks_sharding, vmapped_fun, chunk_size, argnums
        )
    else:
        return HashablePartial(_eval_fun_in_chunks, vmapped_fun, chunk_size, argnums)


def _parse_in_axes(in_axes):
    if isinstance(in_axes, int):
        in_axes = (in_axes,)

    if not set(in_axes).issubset((0, None)):
        raise NotImplementedError("Only in_axes 0/None are currently supported")

    argnums = tuple(
        map(lambda ix: ix[0], filter(lambda ix: ix[1] is not None, enumerate(in_axes)))
    )
    return in_axes, argnums


def apply_chunked(
    f: Callable,
    in_axes=0,
    *,
    chunk_size: int | None,
    axis_0_is_sharded: bool = config.netket_experimental_sharding,  # type: ignore[attr-defined]
) -> Callable:
    """
    Takes an implicitly vmapped function over the axis 0 and uses scan to
    do the computations in smaller chunks over the 0-th axis of all input arguments.

    For this to work, the function `f` should be `vectorized` along the `in_axes`
    of the arguments. This means that the function `f` should respect the following
    condition:

    .. code-block:: python

        assert f(x) == jnp.concatenate([f(x_i) for x_i in x], axis=0)

    which is automatically satisfied if `f` is obtained by vmapping a function,
    such as:

    .. code-block:: python

        f = jax.vmap(f_orig)

    .. note::
        If netket_experimental_sharding is enabled, this function assumes that chunked in_axes are sharded by default.
        This can be overridden by specifying axis_0_is_sharded=False.

    Args:
        f: A function that satisfies the condition above
        in_axes: The axes that should be scanned along. Only supports `0` or `None`
        chunk_size: The maximum size of the chunks to be used. If it is `None`, chunking
            is disabled
        axis_0_is_sharded: specifies if axis 0 of the arrays scanned is sharded among multiple devices,
            The function is then computed in chunks of size chunk_size on every device.
            Defaults True if config.netket_experimental_sharding, oterhwise defaults to False.

    """
    _, argnums = _parse_in_axes(in_axes)
    return _chunk_vmapped_function(f, chunk_size, argnums, axis_0_is_sharded)


def vmap_chunked(
    f: Callable,
    in_axes=0,
    *,
    chunk_size: int | None,
    axis_0_is_sharded: bool = config.netket_experimental_sharding,  # type: ignore[attr-defined]
) -> Callable:
    """
    Behaves like jax.vmap but uses scan to chunk the computations in smaller chunks.

    This function is essentially equivalent to:

    .. code-block:: python

        nk.jax.apply_chunked(jax.vmap(f, in_axes), in_axes, chunk_size)

    Some limitations to `in_axes` apply.

    .. note::
        If netket_experimental_sharding is enabled, this function assumes that chunked in_axes are sharded by default.
        This can be overridden by specifying axis_0_is_sharded=False.

    Args:
        f: The function to be vectorised.
        in_axes: The axes that should be scanned along. Only supports `0` or `None`
        chunk_size: The maximum size of the chunks to be used. If it is `None`, chunking
            is disabled
        axis_0_is_sharded: specifies if axis 0 of the arrays scanned is sharded among multiple devices,
            The function is then computed in chunks of size chunk_size on every device.
            Defaults True if config.netket_experimental_sharding, oterhwise defaults to False.

    Returns:
        A vectorised and chunked function
    """
    in_axes, argnums = _parse_in_axes(in_axes)
    vmapped_fun = jax.vmap(f, in_axes=in_axes)
    return _chunk_vmapped_function(vmapped_fun, chunk_size, argnums, axis_0_is_sharded)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, TYPE_CHECKING
from collections.abc import Callable

import sys

import orjson

import numpy as np

import jax

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

if TYPE_CHECKING:
    from .history import History
else:
    History = Any


def _format_key(k):
    if hasattr(k, "key"):
        return str(k.key)
    elif hasattr(k, "idx"):
        return str(k.idx)
    elif hasattr(k, "name"):
        return str(k, k.name)
    else:
        return str(k)  # fallback


class HistoryDict:
    """A class that behaves like a dictionary, mapping strings to History instances or nested
    HistoryDict instances.

    This class behaves to all effects as a dictionary at runtime, but it allows us to specify
    custom serialization and deserialization routines, which are used to save and load the data.

    .. note::

        This class can be used to deserialize a dictionary of histories, such as the standard
        log files obtained from :class:`~netket.logging.RuntimeLog` and :class:`~netket.logging.JsonLog`.


    .. warn::

        As of november 2024, this class is to be considered an experimental implementation detail,
        used to play well with checkpointing. Do not use directly unless you know what you are doing.

    """

    def __init__(self, *args, **kwargs):
        """
        Create a new HistoryDict instance from a dictionary of histories.
        """
        data = dict(*args, **kwargs)
        # Do not nest HistoryDict inside Historydict!
        for k, d in data.items():
            if isinstance(d, HistoryDict):
                data[k] = d.to_dict()
        self._data = data

    def __setitem__(self, key: str, value: History | Self):
        if isinstance(value, HistoryDict):
            value = value.to_dict()
        self._data[key] = value

    def __getitem__(self, key: str, *, wrap_dicts=True) -> History | Self:
        if key not in self._data and "/" in key:
            keys = key.split("/")
            val = self._data
            for subkey in keys:
                val = val[subkey]  # Traverse through subdictionaries
            if wrap_dicts and isinstance(val, dict):
                return HistoryDict(val)
            return val
        val = self._data[key]
        if wrap_dicts and isinstance(val, dict):
            return HistoryDict(val)
        return val

    def __repr__(self):
        dk, _ = jax.tree_util.tree_flatten_with_path(self._data)
        if len(dk) == 0:
            return "HistoryDict({})"

        _repr_str = f"HistoryDict with {len(dk)} elements:"
        for keys, val in dk:
            path = "/".join([_format_key(k) for k in keys])
            _repr_str += f"\n\t'{path}' -> " + str(val)

        return _repr_str

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    def keys(self):
        return self._data.keys()

    def items(self):
        return self._data.items()

    def values(self):
        return self._data.values()

    def get(self, key: str, default=None):
        """
        Get a value from the dictionary, with a default value if the key is not present.

        If the value is a dictionary, it is wrapped in a HistoryDict instance.

        Args:
            key: The key to retrieve.
            default: The default value to return if the key is not present.
        """
        res = self._data.get(key, default)
        if isinstance(res, dict):
            return HistoryDict(res)
        return res

    def to_dict(self) -> dict[str, dict | np.ndarray]:
        """
        Convert the HistoryDict to normal dictionary, with all nested HistoryDict instances.
        """
        return {k: self.__getitem__(k, wrap_dicts=False) for k in self._data.keys()}

    @classmethod
    def from_file(cls, fname: str) -> Self:
        """
        Create an HistoryDict from a text-file containing its serialization.

        Args:
            fname: The name of the file to read.
        """
        with open(fname) as f:
            data = orjson.loads(f.read())

        data = histdict_to_nparray(data)

        def _recompose(hist_dict):
            # hist_dict = {k: np.array(v) for k, v in hist_dict.items()}
            for _, checker, reconstructor in DESERIALIZATION_REGISTRY:
                if checker(hist_dict):
                    return reconstructor(hist_dict)
            raise ValueError("No matching type found for the given dictionary.")

        def _is_leaf(dic):
            if isinstance(dic, dict):
                if "iters" in dic:
                    return True
            return False

        return cls(jax.tree.map(_recompose, data, is_leaf=_is_leaf))

    def _ipython_key_completions_(self):
        return self._data.keys()


# Loading
# A registry to map types to their precedence, checker, and reconstructor functions
DESERIALIZATION_REGISTRY = []


def register_historydict_deserialization_fun(
    checker: Callable,
    reconstructor: Callable,
    precedence: int = 0,
):
    """
    Register a type with a precedence, checker, and reconstructor.

    Args:
        checker: A callable that takes a dictionary and returns True if it matches the type.
        reconstructor: A callable that takes a dictionary and reconstructs the object.
        precedence: An integer indicating the precedence of this type (higher is checked first).
    """
    DESERIALIZATION_REGISTRY.append((precedence, checker, reconstructor))
    # Sort the registry by precedence in descending order
    DESERIALIZATION_REGISTRY.sort(key=lambda x: x[0], reverse=True)


# Checker and reconstructor for History
def is_history(hist_dict):
    return "iters" in hist_dict


def reconstruct_history(hist_dict):
    from netket.utils.history.history import History

    if "Mean" in hist_dict:
        return History(hist_dict, main_value_name="Mean")
    return History(hist_dict)


# Register History with precedence 10
register_historydict_deserialization_fun(
    is_history, reconstruct_history, precedence=-10
)


"""
Json serializer for complex numbers.

Json does not support complex numbers, which are stored as {"real": ..., "imag": ...} dictionaries.
This is a workaround to store complex numbers in json files.
"""


def _is_number_list(x):
    return (
        isinstance(x, list) and len(x) > 0 and isinstance(x[0], (int, float, complex))
    )


def _is_complex_leaf(subtree):
    # Print only if you want to debug
    # print("checking if subtree is leaf", subtree)
    return (
        isinstance(subtree, dict) and "real" in subtree and "imag" in subtree
    ) or _is_number_list(subtree)


def _convert_complex(subtree):
    if isinstance(subtree, dict) and "real" in subtree and "imag" in subtree:
        return np.array(subtree["real"]) + 1j * np.array(subtree["imag"])
    return np.array(subtree)


def histdict_to_nparray(hist_dict):
    """
    Convert the {'real': ..., 'imag': ...} dictionaries to complex numbers.
    """
    return jax.tree.map(_convert_complex, hist_dict, is_leaf=_is_complex_leaf)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax
from jax import numpy as jnp

from netket.hilbert import TensorHilbert
from netket.utils.dispatch import dispatch


@dispatch
def random_state(hilb: TensorHilbert, key, batches: int, *, dtype):
    keys = jax.random.split(key, hilb._n_hilbert_spaces)

    vs = [
        random_state(hilb._hilbert_spaces[i], keys[i], batches, dtype=dtype)
        for i in range(hilb._n_hilbert_spaces)
    ]

    return jnp.concatenate(vs, axis=1)


def _make_subfun(hilb, i, sub_hi):
    def subfun(args):
        key, state, index = args

        # jax.experimental.host_callback.id_print(index, text=f"printing subfun_{i}:")

        sub_state = state[hilb._cum_indices[i] : hilb._cum_sizes[i]]
        new_sub_state, old_val = flip_state_scalar(
            sub_hi, key, sub_state, index - hilb._cum_indices[i]
        )
        new_state = state.at[hilb._cum_indices[i] : hilb._cum_sizes[i]].set(
            new_sub_state
        )
        return new_state, old_val

    return subfun


@dispatch
def flip_state_scalar(hilb: TensorHilbert, key, state, index):
    subfuns = []
    for i, sub_hi in enumerate(hilb._hilbert_spaces):
        subfuns.append(_make_subfun(hilb, i, sub_hi))

    branches = []
    for i in hilb._hilbert_i:
        branches.append(subfuns[i])

    return jax.lax.switch(index, branches, (key, state, index))



# Copyright 2022 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import warnings
from textwrap import dedent

from numbers import Number

import numpy as np

import jax

from netket.utils.types import DType
from netket.operator import DiscreteOperator, Transpose
from netket.operator._pauli_strings.base import _count_of_locations
from netket.hilbert import AbstractHilbert
from netket.utils.numbers import is_scalar, dtype as _dtype
from netket.utils.optional_deps import import_optional_dependency

from netket.hilbert import Fock, SpinOrbitalFermions


from .utils import (
    _convert_terms_to_spin_blocks,
    _canonicalize_input,
    _check_hermitian,
    transpose_term,
    _remove_dict_zeros,
    _verify_input,
    OperatorDict,
    OperatorTermsList,
    OperatorWeightsList,
    _normal_ordering,
    _pair_ordering,
)


class FermionOperator2ndBase(DiscreteOperator):
    r"""
    A fermionic operator in :math:`2^{nd}` quantization.

    This is a base class that only provides symbolic-level manipulation capabilities,
    and must be inherited from a subclass defining indexing methods.
    """

    def __init__(
        self,
        hilbert: AbstractHilbert,
        terms: list[str] | list[list[list[int]]] = None,
        weights: list[float | complex] | None = None,
        constant: Number = 0,
        cutoff: float = 1e-10,
        dtype: DType | None = None,
    ):
        r"""
        Constructs a fermion operator given the single terms (set of
        creation/annihilation operators) in second quantization formalism.

        This class can be initialized in the following form:
        `FermionOperator2nd(hilbert, terms, weights ...)`.
        The term contains pairs of `(idx, dagger)`, where `idx ∈ range(hilbert.size)`
        (it identifies an orbital) and `dagger` is a True/False flag determining if the
        operator is a creation or destruction operator.
        A term of the form :math:`\hat{a}_1^\dagger \hat{a}_2` would take the form
        `((1,1), (2,0))`, where (1,1) represents :math:`\hat{a}_1^\dagger` and (2,0)
        represents :math:`\hat{a}_2`.
        To split up per spin, use the creation and annihilation operators to build the
        operator.

        Args:
            hilbert: hilbert of the resulting FermionOperator2nd object
            terms: single term operators (see
                example below)
            weights: corresponding coefficients of the single term operators
                (defaults to a list of 1)
            constant: constant contribution, corresponding to the
                identity operator * constant (default = 0)
            cutoff: threshold for the weights, if the absolute value of a weight is below the cutoff, it's discarded.
            dtype: The datatype to use for the matrix elements. Defaults to double precision if
                available.

        Returns:
            A FermionOperator2nd object.

        Example:
            Constructs the fermionic hamiltonian in :math:`2^{nd}` quantization
            :math:`(0.5-0.5j)*(a_0^\dagger a_1) + (0.5+0.5j)*(a_2^\dagger a_1)`.

            >>> import netket as nk
            >>> terms, weights = (((0,1),(1,0)),((2,1),(1,0))), (0.5-0.5j,0.5+0.5j)
            >>> hi = nk.hilbert.SpinOrbitalFermions(3)
            >>> op = nk.operator.FermionOperator2nd(hi, terms, weights)
            >>> op
            FermionOperator2nd(hilbert=SpinOrbitalFermions(n_orbitals=3), n_operators=2, dtype=complex128)
            >>> terms = ("0^ 1", "2^ 1")
            >>> op = nk.operator.FermionOperator2nd(hi, terms, weights)
            >>> op
            FermionOperator2nd(hilbert=SpinOrbitalFermions(n_orbitals=3), n_operators=2, dtype=complex128)
            >>> op.hilbert
            SpinOrbitalFermions(n_orbitals=3)
            >>> op.hilbert.size
            3

        """
        super().__init__(hilbert)

        if not np.isscalar(cutoff) or cutoff < 0:
            raise ValueError(f"invalid cutoff in FermionOperator2ndBase: {cutoff}.")
        self._cutoff = cutoff

        # bring terms, weights into consistent form, autopromote dtypes if necessary
        _operators, dtype = _canonicalize_input(
            terms, weights, dtype, cutoff, constant=constant
        )
        _verify_input(hilbert, _operators, raise_error=True)
        self._dtype = dtype

        # we keep the input, in order to be able to add terms later
        self._operators = _operators
        self._initialized = False
        self._is_hermitian = None  # set when requested
        self._max_conn_size = None

    def _reset_caches(self):
        """
        Cleans the internal caches built on the operator.
        """
        self._initialized = False
        self._is_hermitian = None
        self._max_conn_size = None

    def _setup(self, force: bool = False):
        """Analyze the operator strings and precompute arrays for get_conn inference"""
        if jax.process_index() == 0:
            if (
                isinstance(self.hilbert, SpinOrbitalFermions)
                and self.hilbert.n_fermions_per_spin is not None
            ):
                warnings.warn(
                    dedent(
                        """
                        WARNING: Initializing `netket.operator.FermionOperator2nd` for a Hilbert space with a fixed number of fermions.
                        Consider using `netket.experimental.operator.ParticleNumberAndSpinConservingFermioperator2nd` to reduce the number of connected elements and
                        considerably reduce the computational cost.
                        You can convert this operator by calling `netket.experimental.operator.ParticleNumberAndSpinConservingFermioperator2nd.from_fermionoperator2nd`.
                        """
                    ),
                    stacklevel=2,
                )
            elif (
                isinstance(self.hilbert, SpinOrbitalFermions)
                and self.hilbert.n_fermions is not None
            ) or (
                isinstance(self.hilbert, Fock) and self.hilbert.n_particles is not None
            ):
                warnings.warn(
                    dedent(
                        """
                        WARNING: Initializing `netket.operator.FermionOperator2nd` for a Hilbert space with a fixed number of fermions.
                        Consider using `netket.experimental.operator.ParticleNumberConservingFermioperator2nd` to reduce the number of connected elements and
                        considerably reduce the computational cost.
                        You can convert this operator by calling `netket.experimental.operator.ParticleNumberConservingFermioperator2nd.from_fermionoperator2nd`.
                        """
                    ),
                    stacklevel=2,
                )

    @classmethod
    def from_openfermion(
        cls,
        hilbert: AbstractHilbert,
        of_fermion_operator=None,  # : "openfermion.ops.FermionOperator" type
        *,
        n_orbitals: int | None = None,
        convert_spin_blocks: bool = False,
        cutoff: float = 1e-10,
        dtype: DType = None,
    ) -> "FermionOperator2ndBase":
        r"""
        Converts an openfermion FermionOperator into a netket FermionOperator2nd.

        The hilbert first argument can be dropped, see __init__ for details and default
        value.
        Warning: convention of openfermion.hamiltonians is different from ours: instead
        of strong spin components as subsequent hilbert state outputs (i.e. the 1/2 spin
        components of spin-orbit i are stored in locations (2*i, 2*i+1)), we concatenate
        blocks of definite spin (i.e. locations (i, n_orbitals+i)).

        Args:
            hilbert: (optional) hilbert of the resulting FermionOperator2nd object
            of_fermion_operator (openfermion.ops.FermionOperator):
                `FermionOperator object <https://quantumai.google/reference/python/openfermion/ops/FermionOperator>`_ .
                More information about those objects can be found in
                `OpenFermion's documentation <https://quantumai.google/reference/python/openfermion>`_
            n_orbitals: (optional) total number of orbitals in the system, default
                None means inferring it from the FermionOperator2nd. Argument is
                ignored when hilbert is given.
            convert_spin_blocks: whether or not we need to convert the FermionOperator
                to our convention. Only works if hilbert is provided and if it has
                spin != 0
            cutoff: threshold for the weights, if the absolute value of a weight is below the cutoff, it's discarded.

        """
        openfermion = import_optional_dependency(
            "openfermion", descr="from_openfermion"
        )
        FermionOperator = openfermion.ops.FermionOperator

        if hilbert is None:
            raise ValueError(
                "The first argument `from_openfermion` must either be an "
                "openfermion operator or an Hilbert space, followed by "
                "an openfermion operator"
            )

        if not isinstance(hilbert, AbstractHilbert):
            # if first argument is not Hilbert, then shift all arguments by one
            hilbert, of_fermion_operator = None, hilbert

        if not isinstance(of_fermion_operator, FermionOperator):  # pragma: no cover
            raise NotImplementedError()

        if convert_spin_blocks and hilbert is None:
            raise ValueError("if convert_spin_blocks, the hilbert must be specified")

        terms = list(of_fermion_operator.terms.keys())
        weights = list(of_fermion_operator.terms.values())

        if hilbert is not None:
            # no warning, just overwrite
            n_orbitals = hilbert.n_orbitals

            if convert_spin_blocks:
                if not hasattr(hilbert, "spin") or hilbert.spin is None:
                    raise ValueError(
                        f"cannot convert spin blocks for hilbert space {hilbert} without spin"
                    )
                n_spin = hilbert._n_spin_states
                terms = _convert_terms_to_spin_blocks(terms, n_orbitals, n_spin)
        if n_orbitals is None:
            # we always start counting from 0, so we only determine the maximum location
            n_orbitals = _count_of_locations(of_fermion_operator)
        if hilbert is None:
            hilbert = SpinOrbitalFermions(n_orbitals)  # no spin splitup assumed

        return cls(hilbert, terms, weights=weights, cutoff=cutoff, dtype=dtype)

    def __repr__(self):
        return (
            f"{type(self).__name__}(hilbert={self.hilbert}, "
            f"n_operators={len(self._operators)}, dtype={self.dtype})"
        )

    def reduce(self, order: bool = True, inplace: bool = True, cutoff: float = None):
        """
        Prunes the operator by removing all terms with zero weights, grouping, and normal ordering (inplace).

        Args:
            order: Whether to normal order the operator.
            inplace: Whether to change the current object in place.
            cutoff: Optional cutoff for the weights.
        """
        if cutoff is None:
            cutoff = self._cutoff

        operators = self._operators
        terms, weights = list(operators.keys()), list(operators.values())

        if order:
            terms, weights = _normal_ordering(terms, weights)

        obj = self if inplace else self.copy()
        obj._operators, _ = _canonicalize_input(terms, weights, self.dtype, cutoff)
        return obj

    @property
    def dtype(self) -> DType:
        """The dtype of the operator's matrix elements ⟨σ|Ô|σ'⟩."""
        return self._dtype

    @property
    def terms(self) -> OperatorTermsList:
        """Returns the list of all terms in the tuple notation."""
        return list(self._operators.keys())

    @property
    def weights(self) -> OperatorWeightsList:
        """Returns the list of the weights correspoding to the operator terms."""
        return list(self._operators.values())

    @property
    def operators(self) -> OperatorDict:
        """Returns a dictionary with (term, weight) key-value pairs, with terms in tuple notation.

        The constant diagonal shift of the operator is associated with a term represented as an
        empty tuple.
        """
        return self._operators

    @property
    def cutoff(self) -> float:
        """Threshold for the weights, if the absolute value of a weight is below the
        cutoff, it's discarded.
        """
        return self._cutoff

    def copy(self, *, dtype: DType | None = None, cutoff=None):
        """
        Creates a deep copy of this operator, potentially changing the dtype of the
        operator and internal arrays.

        Args:
            dtype: Optional new dtype. Must be compatible with the current dtype.
            cutoff: Optional new cutoff.

        Returns:
            An identical operator that does not reference the internal arrays of
            the original one.
        """
        if dtype is None:
            dtype = self.dtype
        if not np.can_cast(self.dtype, dtype, casting="same_kind"):
            raise ValueError(f"Cannot cast {self.dtype} to {dtype}")
        if cutoff is None:
            cutoff = self._cutoff

        op = type(self)(self.hilbert, cutoff=self._cutoff, dtype=dtype)

        if dtype == self.dtype:
            operators_new = self._operators.copy()
        else:
            operators_new = {
                k: np.array(v, dtype=dtype) for k, v in self._operators.items()
            }

        op._operators = operators_new
        return op

    def _remove_zeros(self, cutoff: float = None):
        """Reduce the number of operators by removing unnecessary zeros"""
        if cutoff is None:
            cutoff = self._cutoff
        op = type(self)(self.hilbert, cutoff=cutoff, dtype=self.dtype)
        op._operators = _remove_dict_zeros(self._operators, cutoff)
        return op

    @property
    def max_conn_size(self) -> int:
        """The maximum number of non zero ⟨x|O|x'⟩ for every x."""
        if self._max_conn_size is None:
            self._setup()
        return self._max_conn_size

    @property
    def is_hermitian(self) -> bool:
        """Returns true if this operator is hermitian."""
        if self._is_hermitian is None:  # only compute when needed, is expensive
            terms = list(self._operators.keys())
            weights = list(self._operators.values())
            cutoff = self._cutoff
            self._is_hermitian = _check_hermitian(terms, weights, cutoff)
        return self._is_hermitian

    def operator_string(self) -> str:
        """Return a readable string describing all the operator terms"""
        op_string = []
        for term, weight in self._operators.items():
            s = []
            for idx, dag in term:
                dag_string = "^" if bool(dag) else ""
                s.append(f"{int(idx)}{dag_string}")
            s = " ".join(s)
            s = f"{weight} [{s}]"
            op_string.append(s)
        return " +\n".join(op_string)

    def _op__imatmul__(self, other):
        if not isinstance(other, FermionOperator2ndBase):  # pragma: no cover
            return NotImplemented
        if not self.hilbert == other.hilbert:
            raise ValueError(
                "Can only multiply identical hilbert spaces (got A@B, "
                f"A={self.hilbert}, B={other.hilbert})"
            )
        if not np.can_cast(_dtype(other), self.dtype, casting="same_kind"):
            raise ValueError(
                f"Cannot multiply inplace operator with dtype {type(other)} "
                f"to operator with dtype {self.dtype}"
            )

        new_operators = {}
        for t, w in self._operators.items():
            for to, wo in other._operators.items():
                # if the last operator of t and the first of to are
                # equal, we have a ...ĉᵢĉᵢ... which is null.
                if len(t) == 0 or len(to) == 0 or t[-1] != to[0]:
                    new_t = t + to
                    new_operators[new_t] = new_operators.get(new_t, 0) + w * wo

        self._operators = new_operators
        self._reset_caches()
        return self

    def _op__matmul__(self, other):
        if not isinstance(other, FermionOperator2ndBase):  # pragma: no cover
            return NotImplemented
        dtype = np.promote_types(self.dtype, other.dtype)
        cutoff = max(self._cutoff, other._cutoff)
        op = self.copy(dtype=dtype, cutoff=cutoff)
        return op._op__imatmul__(other)

    def __radd__(self, other):
        return self.__add__(other)

    def __add__(self, other):
        dtype = np.promote_types(self.dtype, _dtype(other))
        op = self.copy(dtype=dtype)
        return op.__iadd__(other)

    def __iadd__(self, other):
        if is_scalar(other):
            return self + self.__class__(
                self.hilbert, constant=other, dtype=self.dtype, cutoff=self._cutoff
            )
        if not isinstance(other, FermionOperator2ndBase):  # pragma: no cover
            raise NotImplementedError(
                f"In-place addition not implemented for {type(self)} "
                f"and {type(other)}"
            )
        if not self.hilbert == other.hilbert:
            raise ValueError(
                f"Can only add identical hilbert spaces (got A+B, A={self.hilbert}, "
                f"B={other.hilbert})"
            )
        if not np.can_cast(_dtype(other), self.dtype, casting="same_kind"):
            raise ValueError(
                f"Cannot add inplace operator with dtype {type(other)} "
                f"to operator with dtype {self.dtype}"
            )
        self_ops = self._operators
        for t, w in other._operators.items():
            sw = self_ops.get(t, None)
            if sw is None and not np.isclose(w, 0):
                self_ops[t] = w
            elif sw is not None:
                w = sw + w
                if np.isclose(w, 0):
                    del self_ops[t]
                else:
                    self_ops[t] = w
        self._reset_caches()
        return self

    def __sub__(self, other):
        return self + (-other)

    def __rsub__(self, other):
        return other + (-self)

    def __neg__(self):
        return self.__mul__(np.array(-1, dtype=self.dtype))

    def __rmul__(self, scalar):
        return self * scalar

    def __imul__(self, scalar):
        if not is_scalar(scalar):
            # we will overload this as matrix multiplication
            self._op__imatmul__(scalar)
        if not np.can_cast(_dtype(scalar), self.dtype, casting="same_kind"):
            raise ValueError(
                f"Cannot multiply inplace scalar with dtype {type(scalar)} "
                f"to operator with dtype {self.dtype}"
            )
        scalar = np.array(scalar, dtype=self.dtype).item()

        if np.isclose(scalar, 0):
            new_operators = {}
        else:
            new_operators = {o: scalar * v for o, v in self._operators.items()}

        self._operators = new_operators
        self._reset_caches()
        return self

    def __mul__(self, scalar):
        if not is_scalar(scalar):
            # we will overload this as matrix multiplication
            return self._op__matmul__(scalar)
        dtype = np.promote_types(self.dtype, _dtype(scalar))
        op = self.copy(dtype=dtype)
        return op.__imul__(scalar)

    def transpose(self, *, concrete=False):
        r"""Returns the transpose of this operator.

        For real operators, this is equivalent to complex conjugation.
        This implementation always returns the concrete version of the operator.
        """
        if concrete:
            new = type(self)(self.hilbert, dtype=self.dtype, cutoff=self._cutoff)
            # operators (c,c†) are real already. Only conjugate coefficients if needed.
            new._operators = {transpose_term(k): v for k, v in self._operators.items()}
            return new
        else:
            return Transpose(self)

    def conjugate(self, *, concrete=False):
        r"""Returns the complex conjugate of this operator.

        This implementation always returns the concrete version of the operator.
        """
        del concrete

        # if operator is real dtype, then just return a copy
        if not np.issubdtype(self.dtype, np.complexfloating):
            return self.copy()
        else:
            new = type(self)(self.hilbert, dtype=self.dtype, cutoff=self._cutoff)
            # operators (c,c†) are real already. Only conjugate coefficients if needed.
            new._operators = {k: np.conjugate(v) for k, v in self._operators.items()}
            return new

    def to_normal_order(self):
        """Reoder the operators to normal order.
        Normal ordering corresponds to placing creating operators on the left and annihilation on the right.
        Then, it places the highest index on the left and lowest index on the right
        In this ordering, we make sure to account for the anti-commutation of operators.
        `Normal ordering documentation <https://en.wikipedia.org/wiki/Normal_order#Fermions>`_
        """
        terms, weights = _normal_ordering(self.terms, self.weights)
        new = type(self)(
            self.hilbert,
            dtype=self.dtype,
        )
        new._operators, _ = _canonicalize_input(
            terms, weights, self.dtype, self._cutoff
        )
        new._cutoff = self._cutoff
        new.reduce(order=False)  # already ordered
        return new

    def to_pair_order(self):
        """Reoder the operators to pair order.
        Pair ordering corresponds to placing first the highest indices on the right,
        and then making sure the creation operators are on the left and annihilation on the right.
        In this ordering, we make sure to account for the anti-commutation of operators.
        """
        terms, weights = _pair_ordering(self.terms, self.weights)
        new = type(self)(
            self.hilbert,
            dtype=self.dtype,
        )
        new._operators, _ = _canonicalize_input(
            terms, weights, self.dtype, self._cutoff
        )
        new._cutoff = self._cutoff
        new.reduce(order=False)  # already ordered
        return new



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Sequence

import numpy as np
import igraph

from netket.utils.deprecation import warn_deprecation
from netket.utils.group import Permutation, PermutationGroup
from .abstract_graph import AbstractGraph, Edge, ColoredEdge, EdgeSequence


class Graph(AbstractGraph):
    """
    A simple implementation of Graph based on an external graph library.

    The underlying implementation is based on igraph and supports conversion to
    networkx, but this is an implementation detail and could be changed in the future.
    """

    # Initialization
    # ------------------------------------------------------------------------
    def __init__(
        self,
        edges: Sequence[Edge] | Sequence[ColoredEdge],
        n_nodes: int | None = None,
    ):
        """
        Construct the a graph starting from a list of edges and optionally a given
        number of nodes.

        Args:
            edges: list of (undirected) edges
            n_nodes: number of nodes. Can be used to specify the number vertices in the
                graph if not all vertices appear in an edge.

        """
        edges, colors = self._clean_edges(edges)
        if n_nodes is None:
            if len(edges) > 0:
                n_nodes = max(max(e) for e in edges) + 1
            else:
                n_nodes = 0
        graph = igraph.Graph(directed=False)
        graph.add_vertices(n_nodes)
        graph.add_edges(edges, attributes={"color": colors})

        self._igraph = graph
        self._automorphisms = None

    @staticmethod
    def _clean_edges(edges):
        """Validate and normalize edges argument."""
        if not isinstance(edges, Sequence):
            raise TypeError("edges must be a sequence.")
        if len(edges) == 0:
            return edges, []

        e0 = edges[0]
        if not isinstance(e0, Sequence) or len(e0) not in (2, 3):
            raise ValueError(
                "Edges must be tuple of length 2 (or 3 for colored edges)."
            )
        if not all(len(e) == len(e0) for e in edges):
            raise ValueError("Either all or none of the edges need to specify a color.")

        if len(e0) == 2:
            return edges, [0] * len(edges)
        else:
            return [(v, w) for (v, w, _) in edges], [c for (*_, c) in edges]

    # Conversion
    # ------------------------------------------------------------------------
    @classmethod
    def from_igraph(cls, graph: igraph.Graph) -> "Graph":
        """
        Creates a new Graph instance from an igraph.Graph instance.
        """
        if graph.is_directed():
            raise ValueError("Directed graphs are not currently supported.")
        self = cls.__new__(cls)
        self._igraph = graph.copy()
        self._automorphisms = None

        if "color" not in self._igraph.edge_attributes():
            self._igraph.es.set_attribute_values("color", [0] * self._igraph.ecount())
        else:
            if not all(isinstance(c, int) for c in self.edge_colors):
                raise ValueError(
                    "graph has 'color' edge attributes, but not all colors are integers."
                )

        return self

    @classmethod
    def from_networkx(cls, graph) -> "Graph":
        """
        Creates a new Graph instance from a networkx graph.
        """
        ig = igraph.Graph.from_networkx(graph)
        return cls.from_igraph(ig)

    def to_igraph(self):
        """
        Returns a copy of this graph as an igraph.Graph instance.
        """
        return self._igraph.copy()

    def to_networkx(self):
        """
        Returns a copy of this graph as an igraph.Graph instance.
        This method requires networkx to be installed.
        """
        return self._igraph.to_networkx()

    # Graph properties
    # ------------------------------------------------------------------------
    def adjacency_list(self) -> list[list]:
        return self._igraph.get_adjlist()

    def is_connected(self) -> bool:
        return self._igraph.is_connected()

    def is_bipartite(self) -> bool:
        return self._igraph.is_bipartite()

    @property
    def n_nodes(self) -> int:
        r"""The number of nodes (or vertices) in the graph"""
        return self._igraph.vcount()

    @property
    def n_edges(self):
        r"""The number of edges in the graph."""
        return self._igraph.ecount()

    def nodes(self) -> Sequence[int]:
        return range(self._igraph.vcount())

    def edges(
        self,
        color=None,
        *,
        return_color: bool = False,
        filter_color: int | None = None,
    ) -> EdgeSequence:
        if color is not None:
            warn_deprecation(
                "The color option has been split into return_color and filter_color."
            )
            # need to check for bool first, because bool is a subclass of int
            if isinstance(color, bool):
                return_color = color
            elif isinstance(color, int):
                filter_color = color
            else:
                raise TypeError("Incorrect type for 'color'")

        if not return_color and filter_color is None:
            return self._igraph.get_edgelist()

        edges_with_color = zip(self._igraph.get_edgelist(), self.edge_colors)
        if filter_color is not None:
            edges_with_color = filter(lambda x: x[1] == filter_color, edges_with_color)
        if return_color:
            return [(*e, c) for (e, c) in edges_with_color]
        else:
            return [e for (e, _) in edges_with_color]

    @property
    def edge_colors(self) -> Sequence[int]:
        r"""Sequence of edge colors, in the order of the edges returned by
        :code:`self.edges`."""
        if self.n_edges > 0:
            return self._igraph.es.get_attribute_values("color")
        else:
            return []

    # Graph algorithms
    # ------------------------------------------------------------------------
    def distances(self) -> list[list]:
        return np.array(self._igraph.distances())

    def _compute_automorphisms(self):
        """
        Compute the graph autmorphisms of this graph.
        """
        colors = self.edge_colors
        result = self._igraph.get_isomorphisms_vf2(
            edge_color1=colors, edge_color2=colors
        )

        # sort them s.t. the identity comes first
        result = np.unique(result, axis=0).tolist()
        result = PermutationGroup([Permutation(i) for i in result], self.n_nodes)
        return result

    # TODO turn into a struct.property_cached?
    def automorphisms(self) -> PermutationGroup:
        if self._automorphisms is None:
            self._automorphisms = self._compute_automorphisms()
        return self._automorphisms

    # Output and drawing
    # ------------------------------------------------------------------------
    def __repr__(self):
        return "{}(n_nodes={}, n_edges={})".format(
            str(type(self)).split(".")[-1][:-2], self.n_nodes, self.n_edges
        )


def Edgeless(n_nodes: int) -> Graph:
    """
    Construct a set graph (collection of unconnected vertices).

    Args:
        nodes: An integer number of nodes or a list of ints that index nodes of a graph.

    Example:
        >>> import netket
        >>> g=netket.graph.Edgeless([0,1,2,3])
        >>> print(g.n_nodes)
        4
        >>> print(g.n_edges)
        0
    """
    return Graph([], n_nodes)


def DoubledGraph(graph: AbstractGraph) -> Graph:
    """
    DoubledGraph(graph)

    Constructs a DoubledGraph representing the doubled hilbert space of a density
    operator. The resulting graph is composed of two disjoint sub-graphs identical
    to the input.

    Args:
        graph: The graph to double
    """

    dedges = list(graph.edges())
    n_v = graph.n_nodes

    dnodes = 2 * n_v
    dedges += [(edge[0] + n_v, edge[1] + n_v) for edge in graph.edges()]

    return Graph(n_nodes=dnodes, edges=dedges)


def disjoint_union(*graphs: tuple[Graph, ...]) -> Graph:
    """
    Returns the ordered disjoint union of the input graphs.

    The ordering of the nodes is the same as the input order, so if
    the I-th input graph has :math:`n_i` nodes, the first `[0, n_1[` nodes
    will correspond to the first graph, the nodes between :math:`[n_1, n_1+n2[`
    will correspond to the second graph, and so on and so forth.

    Args:
       *graphs: A variable number of NetKet graphs (or lattices).

    Output:
        A NetKet Graph object.
    """
    if len(graphs) == 0:
        raise ValueError("Must provide at least one graph")
    elif len(graphs) == 1:
        return graphs[0]
    else:
        return Graph.from_igraph(igraph.disjoint_union([g._igraph for g in graphs]))



# Copyright 2023-2024 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc


import jax

from netket.utils import struct


class DiscreteHilbertConstraint(struct.Pytree):
    r"""
    Protocol to define an Abstract Constraint for a discete Hilbert space.

    To define a customized constraint, you must subclass this class and at least implement the
    :code:`__call__` method. The :code:`__call__` method should take as input a matrix encoding a batch of
    configurations, and return a vector of booleans specifying whether they are valid configurations
    or not.

    The :code:`__call__` method must be :code:`jax.jit`-able. If you cannot make it jax-jittable, you can implement
    it in numba/python and wrap it into a :func:`jax.pure_callback` to make it compatible with jax.

    The callback should be hashable and comparable with itself, which means it must implement :code:`__hash__` and :code:`__eq__`.
    By default, the :code:`__hash__` method is implemented by the `id` of the object, which is unique for each object,
    which will work but might lead to more recompilations in jax. If you can, you should implement a custom :code:`__hash__`

    Example:

        The following example shows a class that implements a simple constraint checking that the total sum of the
        elements in the configuration is equal to a given value. The example shows how to implement the :code:`__call__` method
        and the :code:`__hash__` and :code:`__eq__` methods.

        .. code-block:: python

            import netket as nk
            from netket.utils import struct

            import jax; import jax.numpy as jnp

            class SumConstraint(nk.hilbert.constraint.DiscreteHilbertConstraint):
                # A simple constraint checking that the total sum of the elements
                # in the configuration is equal to a given value.

                # The value must be set as a pytree_node=False field, meaning
                # that it is a constant and changes to this value represent different
                # constraints.
                total_sum : float = struct.field(pytree_node=False)

                def __init__(self, total_sum):
                    self.total_sum = total_sum

                def __call__(self, x):
                    # Makes it jax-compatible
                    return jnp.sum(x, axis=-1) == self.total_sum

                def __hash__(self):
                    return hash(("SumConstraint", self.total_sum))

                def __eq__(self, other):
                    if isinstance(other, SumConstraint):
                        return self.total_sum == other.total_sum
                    return False

    Example:

        The following example shows how to implement the same function as above, but using a pure python function and
        a :func:`jax.pure_callback` to make it compatible with jax.

        .. code-block:: python

            import jax
            import jax.numpy as jnp
            import numpy as np

            import netket as nk
            from netket.utils import struct

            class SumConstraintPy(nk.hilbert.constraint.DiscreteHilbertConstraint):
                # A simple constraint checking that the total sum of the elements
                # in the configuration is equal to a given value.

                total_sum : float = struct.field(pytree_node=False)

                def __init__(self, total_sum):
                    self.total_sum = total_sum

                def __call__(self, x):
                    return jax.pure_callback(self._call_py,
                                            (jax.ShapeDtypeStruct(x.shape[:-1], bool)),
                                            x,
                                            vmap_method="expand_dims")

                def _call_py(self, x):
                    # Not Jax compatible
                    return np.sum(x, axis=-1) == self.total_sum

                def __hash__(self):
                    return hash(("SumConstraintPy", self.total_sum))

                def __eq__(self, other):
                    if isinstance(other, SumConstraintPy):
                        return self.total_sum == other.total_sum
                    return False

    """

    @abc.abstractmethod
    def __call__(self, x: jax.Array) -> jax.Array:
        """
        This function should take as input a matrix encoding a batch of configurations,
        and return a vector of booleans specifying whether they are valid configurations of
        the Hilbert space or not.

        Args:
            x: 2D matrix.
        """



# utilities to convert from / to FermionOperator2nd/FermionOperator2ndJax


import numpy as np

import jax.numpy as jnp

from netket.jax import COOArray
from netket.utils.types import Array

from .._normal_order_utils import (
    split_spin_sectors,
    OperatorArrayDict,
    SpinOperatorArrayDict,
)

from netket.operator._fermion2nd.utils import OperatorTermsList, OperatorWeightsList


# TODO merge this with fermionoperator2nd prepare_terms_list
def fermiop_to_pnc_format_helper(
    terms: OperatorTermsList, weights: OperatorWeightsList
) -> OperatorArrayDict:
    r"""
    helper function to convert OperatorTermsList and OperatorWeightsList to OperatorArrayDict

    Args:
        terms: terms as specified in FermionOperator2nd/FermionOperator2ndJax
        weights: a list of weights
    Returns:
        a dictionary {k: (sites, daggers, weights)}
        where for every set of operators of length k
            sites: (n_terms, k) matrix containing the indices of the c/c^\dagger operators
            daggers: (n_terms,k), matrix storing c/c^\dagger as 0/1
            weights: (n_terms,) vector of corresponding weights
    """
    out = {}
    for t, w in zip(terms, weights):
        if len(t) == 0:  # constant
            out[0] = (
                np.zeros((1, 0), dtype=np.int32),
                np.zeros((1, 0), dtype=np.int8),
                np.array([w]),
            )
        else:
            sites, daggers = np.array(t).T
            l = len(daggers)
            assert l % 2 == 0
            assert 2 * daggers.sum() == l
            tl, dl, wl = out.get(l, ([], [], []))
            out[l] = tl + [sites,], dl + [daggers,], wl + [w,]  # fmt: skip
    return {
        k: (
            jnp.array(v[0], dtype=np.int32),
            jnp.array(v[1], dtype=np.int8),
            jnp.array(v[2]),
        )
        for k, v in out.items()
    }


def fermiop_to_pnc_format_spin_helper(
    terms: OperatorTermsList,
    weights: OperatorWeightsList,
    n_orbitals: int,
    n_spin_subsectors: int,
) -> SpinOperatorArrayDict:
    r"""
    convert OperatorTermsList and OperatorWeightsList to SpinOperatorArrayDict
    """
    # output: { size : (sites, sectors, daggers, weights) }
    return split_spin_sectors(
        fermiop_to_pnc_format_helper(terms, weights),
        n_orbitals,
        n_spin_subsectors,
    )


def pnc_format_to_fermiop_helper(
    index_array: Array, create_array: Array, weight_array: Array
) -> tuple[OperatorTermsList, OperatorWeightsList]:
    r"""
    convert OperatorArrayTerms to OperatorTermsList and OperatorWeightsList
    """
    if index_array is None:  # diagonal
        if weight_array.ndim == 0:  # const
            return np.array([()], dtype=np.int32), np.array(weight_array)
        else:
            if not isinstance(weight_array, COOArray):
                weight_array = COOArray.fromdense(weight_array)
            destr = np.array(weight_array.coords.T)
            weights = np.array(weight_array.data)
            sites = np.concatenate([destr, destr], axis=-1)
    else:
        if index_array.ndim == 0:  # const
            return np.array([()], dtype=np.int32), np.array(weight_array)
        if not isinstance(index_array, COOArray):
            index_array = COOArray.fromdense(index_array)
        ind = np.array(index_array.data)
        destr = np.array(index_array.coords.T[:, None, :])
        create = create_array[ind]
        destr = np.broadcast_to(destr, create.shape)
        weights = weight_array[ind]
        sites = np.concatenate([destr, create], axis=-1)

    # flatten
    weights = weights.reshape(-1)
    sites = sites.reshape(-1, sites.shape[-1])

    daggers = np.zeros_like(sites)
    daggers[:, : daggers.shape[1] // 2] = 1
    terms = np.concatenate([sites[..., None], daggers[..., None]], axis=-1)

    return terms, weights



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, TYPE_CHECKING
from numbers import Number

from netket.utils.optional_deps import import_optional_dependency

from .base import AbstractLog

if TYPE_CHECKING:
    import tensorboardX
    from netket.vqs import VariationalState


def tree_log(tree, root, data):
    """
    Maps all elements in tree, recursively calling tree_log with a new root string,
    and when it reaches leaves pushes (string, leave) tuples to data.

    Args:
        tree: a pytree where the leaf nodes contain data
        root: the root of the tags used to log to tensorboard
        data: a container modified in place

    """
    if tree is None:
        return
    elif isinstance(tree, list):
        for i, val in enumerate(tree):
            tree_log(val, root + f"/{i}", data)

    elif isinstance(tree, list) and hasattr(tree, "_fields"):
        for key in tree._fields:  # type: ignore
            tree_log(getattr(tree, key), root + f"/{key}", data)

    elif isinstance(tree, tuple):
        for i, val in enumerate(tree):
            tree_log(val, root + f"/{i}", data)

    elif isinstance(tree, dict):
        for key, value in tree.items():
            tree_log(value, root + f"/{key}", data)  # noqa: F722

    elif hasattr(tree, "to_compound"):
        tree_log(tree.to_compound()[1], root, data)  # noqa: F722  # type: ignore

    elif hasattr(tree, "to_dict"):
        tree_log(tree.to_dict(), root, data)  # noqa: F722  # type: ignore

    elif isinstance(tree, complex):
        tree_log(tree.real, root + "/re", data)  # noqa: F722
        tree_log(tree.imag, root + "/im", data)  # noqa: F722

    else:
        data.append((root, tree))


class TensorBoardLog(AbstractLog):
    """
    Creates a tensorboard logger using tensorboardX's summarywriter.

    Refer to its documentation for further details

    https://tensorboardx.readthedocs.io/en/latest/tensorboard.html

    TensorBoardX must be installed.

    Args:
        logdir (string): Save directory location. Default is
          runs/**CURRENT_DATETIME_HOSTNAME**, which changes after each run.
          Use hierarchical folder structure to compare
          between runs easily. e.g. pass in 'runs/exp1', 'runs/exp2', etc.
          for each new experiment to compare across them.
        comment (string): Comment logdir suffix appended to the default
          ``logdir``. If ``logdir`` is assigned, this argument has no effect.
        purge_step (int):
          When logging crashes at step :math:`T+X` and restarts at step :math:`T`,
          any events whose global_step larger or equal to :math:`T` will be
          purged and hidden from TensorBoard.
          Note that crashed and resumed experiments should have the same ``logdir``.
        max_queue (int): Size of the queue for pending events and
          summaries before one of the 'add' calls forces a flush to disk.
          Default is ten items.
        flush_secs (int): How often, in seconds, to flush the
          pending events and summaries to disk. Default is every two minutes.
        filename_suffix (string): Suffix added to all event filenames in
          the logdir directory. More details on filename construction in
          tensorboard.summary.writer.event_file_writer.EventFileWriter.
        write_to_disk (boolean):
          If pass `False`, TensorBoardLog will not write to disk.

    Examples:
        Logging optimisation to tensorboard.

        >>> import pytest; pytest.skip("skip automated test of this docstring")
        >>>
        >>> import netket as nk
        >>> # create a summary writer with automatically generated folder name.
        >>> writer = nk.logging.TensorBoardLog()
        >>> # folder location: runs/May04_22-14-54_s-MacBook-Pro.local/
        >>> # create a summary writer using the specified folder name.
        >>> writer = nk.logging.TensorBoardLog("my_experiment")
        >>> # folder location: my_experiment
        >>> # create a summary writer with comment appended.
        >>> writer = nk.logging.TensorBoardLog(comment="LR_0.1_BATCH_16")
        >>> # folder location: runs/May04_22-14-54_s-MacBook-Pro.localLR_0.1_BATCH_16/
    """

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        self._init_args = args
        """Store the args for the lazily initialized SummaryWriter's constructor."""
        self._init_kwargs = kwargs
        """Store the kwargs for the lazily initialized SummaryWriter's constructor."""

        self._writer: tensorboardX.SummaryWriter | None = None
        """Lazily initialized summarywriter constructor"""

        self._old_step = 0

    def _init_tensorboard(self):
        tensorboardX = import_optional_dependency(
            "tensorboardX", descr="TensorBoardLog"
        )

        self._writer = tensorboardX.SummaryWriter(*self._init_args, **self._init_kwargs)

    def __call__(
        self,
        step: int,
        item: dict[str, Any],
        variational_state: "VariationalState | None" = None,
    ):
        if not self._is_master_process:
            return
        if self._writer is None:
            self._init_tensorboard()
        assert self._writer is not None

        data = []
        tree_log(item, "", data)

        for key, val in data:
            if isinstance(val, Number):
                self._writer.add_scalar(key[1:], val, step)

        self._writer.flush()
        self._old_step = step

    def __del__(self):
        self.flush()

    def _flush_log(self):
        if self._writer is not None:
            self._writer.flush()

    def _flush_params(self, _):
        return None

    def flush(self, variational_state=None):
        """
        Writes to file the content of this logger.

        :param machine: optionally also writes the parameters of the machine.
        """
        self._flush_log()

        if variational_state is not None:
            self._flush_params(variational_state)



# utilities to bring the intermediate internal representation into normal order

import numpy as np
from netket.utils.types import Array


# these types contain the same information as the types in netket.operator._fermion2nd.utils
# but use arrays to store the information

OperatorArrayTerms = tuple[Array, Array, Array]
r"""
A sum of n_terms fermionic operators with fixed n_operators number of creation/annihilation operators
A tuple (sites, daggers, weights) where
    sites: An integer array of size n_terms x n_operators containing the indices i
    daggers: An boolean array of size n_terms x n_operators specifying if the operator is creation/destruction
    weights: An array of size n_terms containing the weight of each term

Example:
The operator :math:`1.0 \hat{c}_1^\dagger \hat{c}_2 + 2.0 \hat{c}_3 \hat{c}_1^\dagger` is represented by
:code:`sites = [[1,2],[3,1]], daggers = [[1,0],[0,1]], weights = [1.0, 2.0]``
"""

OperatorArrayDict = dict[int, OperatorArrayTerms]
r"""
A dictionary containing OperatorArrayTerms of different lengths, where the key specifies the lenght,
representing a generic fermionic operator in second quantization.
"""

SpinOperatorArrayTerms = tuple[Array, Array, Array, Array]
r"""
A sum of n_terms fermionic operators with fixed n_operators number of creation/annihilation operators in n_spin_subsectors different spin secors
A tuple(sites, sectors, daggers, weights) where
    sites: An integer array of size n_terms x n_operators containing the indices i
    sectors: An integer array of size n_terms x n_operators containing the spin sector in 0,...,n_spin_subsectors-1
    daggers: An boolean array of size n_terms x n_operators specifying if the operator is creation/destruction
    weights: An array of size n_terms containing the weight of each term

Example:
The operator :math:`1.0 \hat{c}_{1,\downarrow}^\dagger \hat{c}_{2,\downarrow} + 2.0 \hat{c}_{3,\downarrow} \hat{c}_{1,\uparrow}^\dagger` is represented by
:code:`sites = [[1,2],[3,1]], sectors=[[0,0], [0,1]], daggers = [[1,0],[0,1]], weights = [1.0, 2.0]`
where we encode :math:`\downarrow,\uparrow` as :code:`0,1`
"""
SpinOperatorArrayDict = dict[int, SpinOperatorArrayTerms]
r"""
A dictionary containing SpinOperatorArrayTerms of different lengths, where the key specifies the lenght,
representing a generic fermionic operator in second quantization.
Version with spin sectors.
"""


def parity(x: Array):
    """
    compute the parity of a permutation (along axis 1)
    vectorized along all leading batch dimensions

    Args:
        x: a numpy array containing the permutation (e.g. obtained from np.argsort)

    Returns:
        False/True depending if an even/odd number of swaps is needed to bring x into order

    this function is equivalent to
    .. code:: python
      import numpy as np
      from functools import partial
      from sympy.combinatorics.permutations import Permutation
      parity = partial(np.apply_along_axis, lambda x: Permutation(x).parity(), -1)

    Example:
        >>> import numpy as np
        >>> from netket.experimental.operator._normal_order_utils import parity
        >>> print(parity(np.array([0,1,2])))
        False
        >>> print(parity(np.array([0,2,1])))
        True
        >>> print(parity(np.array([2,0,1])))
        False
    """
    # see https://github.com/sympy/sympy/blob/96836db06ba6b78103dd4217db53db31502fb2f6/sympy/combinatorics/permutations.py#L115

    n = x.shape[-1]
    A = x[..., :, None] < x[..., None, :]
    mask = np.tril(np.ones((n, n), dtype=bool))
    mask = np.expand_dims(mask, tuple(range(0, A.ndim - 2)))
    return np.bitwise_xor.reduce(A & mask, axis=(-2, -1))


def _prune(sites: Array, daggers: Array, weights: Array) -> OperatorArrayTerms:
    r"""
    remove ĉᵢĉᵢ and ĉᵢ†ĉᵢ† on the same site i

    Args:
        sites: An integer array of size n_terms x n_operators containing the indices i
        daggers: An boolean array of size n_terms x n_operators specifying if the operator is creation/destruction
        weights: An array of size n_terms containing the weight of each term
    Returns:
        (sites, daggers, weights) where the terms with repeated ĉᵢĉᵢ/ĉᵢ†ĉᵢ† have been removed
    """
    mask = ~((np.diff(daggers, axis=-1) == 0) & (np.diff(sites, axis=-1) == 0)).any(
        axis=-1
    )
    return sites[mask], daggers[mask], weights[mask]


def _move(i: Array, j: Array, x: Array, mask: Array = None) -> Array:
    r"""
    move element i after element j on the last axis (batched)
    use the mask parameter to only selectively move some elements in the batch

    Args:
        i: integer array of shape (...,) with values in 0,...,n-1
        j: integer array of shape (...,) with values in 0,...,n-1
        x: array of shape (..., n)
        mask: boolean array of size (...,)

    Returns:
        an array where the i'th element of x has been moved after the jth element if mask is True,
        else x if mask is False

    Example: if x is a 2-dimensional matrix of shape (m,n) and i,j are vectors of size m then,
    in each row k in 0,...,m-1, this removes the element A[k, i[k]] and inserts it after element A[k, j[k]]
    """
    n = x.shape[-1]
    a = np.arange(n)[None]
    masklr = (a < i) | (a > j)
    maskj = a == j
    mask_middle = ~(masklr | maskj)
    x1 = np.roll(x, -1, axis=-1)
    xi = np.take_along_axis(x, i, 1)
    res = masklr * x + mask_middle * x1 + maskj * xi
    if mask is None:
        return res
    return res * mask + (~mask) * x


def _remove(i: Array, j: Array, x: Array) -> Array:
    r"""
    remove element i and j on the last axis (batched)

    Args:
        i: integer array of shape (...,) with values in 0,...,n-1
        j: integer array of shape (...,) with values in 0,...,n-1
        x: array of shape (..., n)
        mask: boolean array of size (...,)

    Returns:
        an array of shape (..., n-2) where the i'th and jth element of x has been removed

    Example: if x is a 2-dimensional matrix of shape (m,n) and i,j are vectors of size m then,
    in each row k in 0,...,m-1, this removes the elements in A[k, i[k]] and A[k,j[k]]
    resulting in a (m,n-2) matrix
    """
    n = x.shape[-1]
    a = np.arange(n)[None]
    maskl = a < i
    maskr = a > j - 2
    mask_middle = ~(maskl | maskr)
    x1 = np.roll(x, -1, axis=-1)
    x2 = np.roll(x, -2, axis=-1)
    return (maskl * x + mask_middle * x1 + maskr * x2)[..., :-2]


def _move_daggers_left(
    sites: Array, daggers: Array, weights: Array
) -> tuple[OperatorArrayTerms]:
    r"""
    apply the fermionic anticommutation rules to recursively bring a set of
    fermionic operators (all of same length) into normal ordering (daggers to the left)


    Args:
        sites: An integer array of size n_terms x n_operators containing the indices i
        daggers: An boolean array of size n_terms x n_operators specifying if the operator is creation/destruction
        weights: An array of size n_terms containing the weight of each term

    Returns:
        a tuple of ((sites, daggers, weights), (sites, daggers, weights), ...)
        of operators with the same length, length-2, ..., length 0
    """

    # TODO
    # re-implement non-recursively using Wick thm

    n = daggers.shape[-1]
    if n == 0:
        return ((sites, daggers, weights),)

    sites_ = sites
    daggers_ = daggers
    weights_ = weights

    new_sites_smaller = []
    new_daggers_smaller = []
    new_weights_smaller = []

    while True:
        a = np.arange(n)[None]
        # find leftmost c
        i = np.argmin(daggers_, axis=1, keepdims=True)
        # find next c^\dagger
        j = np.argmax((i < a) & daggers_, axis=1, keepdims=True)

        # now move the c after the c^\dagger if necessary

        do_move = j > i
        if ~do_move.any():
            break

        sign = 1 - 2 * (((j - i) * do_move) % 2).ravel()

        si = np.take_along_axis(sites_, i, 1)
        sj = np.take_along_axis(sites_, j, 1)

        new_sites = _move(i, j, sites_, mask=do_move)
        new_daggers = _move(i, j, daggers_, mask=do_move)
        new_weights = weights_ * sign

        same = ((si == sj) & do_move).ravel()
        new_sites2 = _remove(i[same], j[same], sites_[same])
        new_daggers2 = _remove(i[same], j[same], daggers_[same])
        new_weights2 = -(weights_ * sign)[same]
        new_sites2, new_daggers2, new_weights2 = _prune(
            new_sites2, new_daggers2, new_weights2
        )

        if len(new_sites2) > 0:
            new_sites_smaller.append(new_sites2)
            new_daggers_smaller.append(new_daggers2)
            new_weights_smaller.append(new_weights2)

        # set variables for next iteration
        sites_, daggers_, weights_ = _prune(new_sites, new_daggers, new_weights)

    if len(new_sites_smaller) > 0:
        new_sites_smaller = np.concatenate(new_sites_smaller, axis=0)
        new_daggers_smaller = np.concatenate(new_daggers_smaller, axis=0)
        new_weights_smaller = np.concatenate(new_weights_smaller, axis=0)
        # recursion; TODO collapse first and only run once for each size instead?
        return (sites_, daggers_, weights_), *_move_daggers_left(
            new_sites_smaller, new_daggers_smaller, new_weights_smaller
        )
    else:
        return ((sites_, daggers_, weights_),)


def move_daggers_left(t: OperatorArrayDict) -> OperatorArrayDict:
    r"""
    Apply the fermionic anticommutation rules to recursively bring a set of
    fermionic operators into normal ordering (daggers to the left)

    Args:
        t: a dictionary containing strings of different lengths, each stored as a tuple
        (sites, daggers, weights)

    Returns:
        a new dictionary with normal ordered strings

    """
    d = {}
    for sdw in [x for v in t.values() for x in _move_daggers_left(*v)]:
        k = sdw[0].shape[-1]
        if sdw[-1].size > 0:  # not empty
            di = d.pop(k, None)
            if di is None:
                d[k] = sdw
            else:
                d[k] = tuple(np.concatenate([a, b], axis=0) for a, b in zip(di, sdw))
    return d


def _to_desc_order(sites: Array, daggers: Array, weights: Array) -> OperatorArrayTerms:
    r"""
    Reorder operators (all of same length) such that the ones with the larger site index are to the left
    Assumes the operators are already in normal order (daggers to the left).


    Args:
        sites: An integer array of size n_terms x n_operators containing the indices i
        daggers: An boolean array of size n_terms x n_operators specifying if the operator is creation/destruction
        weights: An array of size n_terms containing the weight of each term

    Returns:
        (sites, daggers, weights) with sites in descending order
    """
    n = daggers.shape[-1]
    if n == 0:
        return sites, daggers, weights
    # check min and max do not over/underflow
    # TODO promote to signed / bigger dtype if necessary
    xl = sites.min() - 1
    xr = sites.max() + 1
    assert (xl < sites.min()).all()
    assert (xr > sites.max()).all()

    # minus because we order descending
    s0 = -daggers * xr - (1 - daggers) * sites
    s1 = -daggers * sites - (1 - daggers) * xl

    perm0 = np.argsort(s0, axis=-1)
    perm1 = np.argsort(s1, axis=-1)
    a = np.arange(len(sites))[:, None]
    sites_desc = sites[a, perm0] * (1 - daggers) + sites[a, perm1] * daggers
    weights_desc = weights * (1 - 2 * (parity(perm0) ^ parity(perm1)))

    # TODO also merge duplicates
    return _prune(sites_desc, daggers, weights_desc)


def to_desc_order(t: OperatorArrayDict) -> OperatorArrayDict:
    r"""
    Reorder operators such that the ones with the larger site index are to the left
    Assumes the operators are already in normal order (daggers to the left).

    Args:
        t: a dictionary containing strings of different lengths, each stored as a tuple
        (sites, daggers, weights)

    Returns:
        a new dictionary with strings in descending order
    """
    # TODO sum duplicates
    return {k: _to_desc_order(*v) for k, v in t.items()}


def to_normal_order(t: OperatorArrayDict) -> OperatorArrayDict:
    r"""
    Apply the fermionic anticommutation rules to recursively bring a set of
    fermionic operators into normal ordering (daggers to the left), then
    reorder operators such that the ones with the larger site index are to the left

    Args:
        t: a dictionary containing strings of different lengths, each stored as a tuple
        (sites, daggers, weights)

    Returns:
        a new dictionary with strings in descending normal order
    """
    return to_desc_order(move_daggers_left(t))


def _split_spin_sectors_helper(
    sites: Array,
    daggers: Array,
    weights: Array,
    n_orbitals: int,
    n_spin_subsectors: int,
) -> SpinOperatorArrayTerms:
    n_ops = sites.shape[1]
    if n_ops == 0:
        return sites, np.zeros_like(sites), daggers, weights
    L = np.arange(n_spin_subsectors) * n_orbitals
    R = np.arange(1, n_spin_subsectors + 1) * n_orbitals
    # n_terms x n_ops x n_spin_subsectors
    sectors_mask = (sites[..., None] >= L) & (sites[..., None] < R)
    sectors = np.einsum("...i,i", sectors_mask, np.arange(n_spin_subsectors)).astype(
        np.int32
    )
    sites = sites - sectors * n_orbitals
    return sites, sectors, daggers, weights


def split_spin_sectors(
    d: OperatorArrayDict, n_orbitals: int, n_spin_subsectors: int
) -> SpinOperatorArrayDict:
    r"""
    Split global site indices into spin sector and index within sector

    Args:
        d: { size : (sites, daggers, weights) }
        n_orbitals: number of orbitals (assumed to be the same for each sector)
        n_spin_subsectors: number of spin sectors
    Returns:
        A dictionary { size : (sites, sectors, daggers, weights) }
        where the site indices have been transformed into the spin sectors and index within the sector
    """
    return {
        k: _split_spin_sectors_helper(*v, n_orbitals, n_spin_subsectors)
        for k, v in d.items()
    }


def _merge_spin_sectors_helper(
    sites: Array, sectors: Array, daggers: Array, weights: Array, n_orbitals: int
) -> OperatorArrayTerms:
    return sites + sectors * n_orbitals, daggers, weights


def merge_spin_sectors(d: SpinOperatorArrayDict, n_orbitals) -> OperatorArrayDict:
    r"""
    Merge spin sector and index within sector into global site index

    Args:
        d: { size : (sites, sectors, daggers, weights) }
        n_orbitals: number of orbitals (assumed to be the same for each sector)

    Returns:
        A dictionary { size : (sites, daggers, weights) }
        where the indices and sectors have been transformed into global site indices
    """
    return {k: _merge_spin_sectors_helper(*v, n_orbitals) for k, v in d.items()}


def to_normal_order_sector(
    t: SpinOperatorArrayDict, n_spin_subsectors: int, n_orbitals: int
) -> SpinOperatorArrayDict:
    r"""convert to normal order with higher sector to the left

    Args:
        t: a dictionary containing strings of different lengths, each stored as a tuple
        (sites, sectors, daggers, weights)
        n_spin_subsectors: number of spin sectors
        n_orbitals: number of orbitals (assumed to be the same for each sector)

    Returns:
        a new dictionary with strings in descending normal order

    """
    return split_spin_sectors(
        to_normal_order(merge_spin_sectors(t, n_orbitals)),
        n_orbitals,
        n_spin_subsectors,
    )



# Copyright 2023 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import jax
import jax.numpy as jnp
import numpy as np
import itertools
from scipy.sparse import issparse

from netket.operator import PauliStrings


_pauli_basis = None
pauli_basis_str = list("IXYZ")


def _get_pauli_basis():
    global _pauli_basis
    if _pauli_basis is None:
        # Pauli Matrices: shape (2, 2)
        I = jnp.eye(2)
        X = jnp.array([[0, 1], [1, 0]])
        Y = jnp.array([[0, -1j], [1j, 0]])
        Z = jnp.array([[1, 0], [0, -1]])

        # stacked pauli matrices: shape (4, 2, 2)
        _pauli_basis = jnp.stack([I, X, Y, Z], axis=0)
    return _pauli_basis


@jax.jit
def _tensor_product(mats1, mats2):
    """Get all combinations of Pauli operators between two sets
    via tensorproduct.

    Args:
        mats1: either a (2ⁿ,2ⁿ) representing a single operator or a (N,2ⁿ,2ⁿ)
            tensor encoding N different pauli operators
        mats2: either a (2ᵐ,2ᵐ) representing a single operator or a (M,2ᵐ,2ᵐ)
            tensor encoding M different pauli operators
    Returns
        A (2ⁿ⁺ᵐ, 2ⁿ⁺ᵐ) tensor if both mats1 and mats2 are single paulis. Otherwise
        an (N,M, 2ⁿ⁺ᵐ, 2ⁿ⁺ᵐ) tensor (where N or M dimension might be dropped if
        not there in the input).
    """
    if mats1.ndim == 2 and mats2.ndim == 2:
        return jnp.kron(mats1, mats2)
    elif mats1.ndim == 3:
        return jax.vmap(_tensor_product, in_axes=(0, None))(mats1, mats2)
    elif mats2.ndim == 3:
        return jax.vmap(_tensor_product, in_axes=(None, 0))(mats1, mats2)
    else:
        raise ValueError("Not supposed to get more than 3 dimensions")


def _get_basis_till_n(n=1):
    """Get the basis set till the basis set with at most n non
    identities in the Pauli Strings, which will contain
    4ⁿ elements of size (2ⁿ, 2ⁿ).

    Returns:
        A dictionary with keys [1, 2, ..., n] where every value is
        a basis set encoded into a tensor of shape (4ⁿ, 2ⁿ, 2ⁿ)
        where the first dimension (4ⁿ) is the number of basis elements
        for that number of non-identity paulis, and 2ⁿ is the size
        of those matrices.

        The key represents the size of those matrices.
    """
    if n == 1:
        return {n: _get_pauli_basis()}
    else:
        lower_bases = _get_basis_till_n(n=n - 1)

        # extract the basis for n-1, which is (4ⁿ⁻¹, 2ⁿ, 2ⁿ) tensor
        # compute the (4,4ⁿ⁻¹, 4, 4)
        tp = _tensor_product(_get_pauli_basis(), lower_bases[n - 1])

        # reshape everything to a single set of bases (4ⁿ, ...)
        tp = tp.reshape(-1, *tp.shape[-2:])
        return {**lower_bases, n: tp}


def _make_basis_strings(n_acting):
    """Make the strings corresponding to the tensorproduct of Pauli operators"""
    bs = list(itertools.product(pauli_basis_str, repeat=n_acting))
    bs = list(map(lambda b: "".join(b), bs))
    return np.array(bs)


def _get_paulistrings_till_n(n=1):
    """
    Get the
    """
    return {n_paulis: _make_basis_strings(n_paulis) for n_paulis in range(1, n + 1)}


@jax.jit
def _hilbert_schmidt(A, basis_element):
    """Hilbert schmidt product where the basis elements are hermitian"""
    return jnp.trace(basis_element @ A)


@jax.jit
def _project_matrix(A, basis_set):
    """
    Project matrix onto basis of tensorproduct of Paulis
    according to hilbert-schmidt product
    """
    assert A.ndim == 2
    assert A.shape[0] == A.shape[1]
    size = A.shape[0]
    return jax.vmap(_hilbert_schmidt, in_axes=(None, 0))(A, basis_set) / size


def _create_pauli_string(substr, acting_on, hilbert_size):
    """Create a single PauliString from the operators and indices"""
    s = list("I" * hilbert_size)
    for i, si in zip(acting_on, substr):
        s[i] = si
    s = "".join(list(s))
    return s


def _local_operator_to_pauli_strings(
    op, acting_on, basis_set, basis_set_str, hilbert_size
):
    """Convert a matrix into a sum of Pauli strings"""
    acting_on = np.array(acting_on)
    all_weights = _project_matrix(op, basis_set)
    nz_idxs = np.nonzero(all_weights)[0]
    weights = all_weights[nz_idxs]

    # the next is not efficient
    operators = [basis_set_str[idx] for idx in nz_idxs]
    _str_maker = lambda s: _create_pauli_string(s, acting_on, hilbert_size)
    operators = list(map(_str_maker, operators))
    return list(operators), list(weights)


def _convert_to_dense(m):
    return m.todense() if issparse(m) else m


def local_operators_to_pauli_strings(
    hilbert, operators, acting_on, constant, dtype, cls=PauliStrings
):
    """Convert a LocalOperator into PauliStrings

    Args:
        local_op: LocalOperator

    Returns:
        PauliStrings
    """

    if any(d != 2 for d in hilbert.shape):
        raise TypeError(
            "Cannot convert to Pauli strings operators defined on hilbert spaces "
            "with local dimension != 2"
        )

    pauli_strings = []
    weights = []

    if len(operators) > 0:
        # maximum number of non-identity operators
        n_max = max(list(map(len, acting_on)))

        # make the basis and the string representation
        bases = _get_basis_till_n(n=n_max)
        bases_strs = _get_paulistrings_till_n(n=n_max)

        # loop trough operators and convert to Pauli strings by projecting
        for op, act in zip(operators, acting_on):
            # convert to dense-numpy as we later use jax which does not support sparse.
            op = _convert_to_dense(op)

            n_paulis = len(act)

            op_list, w_list = _local_operator_to_pauli_strings(
                op, act, bases[n_paulis], bases_strs[n_paulis], hilbert.size
            )
            pauli_strings += op_list
            weights += w_list

    # add the constant if it's not zero
    if np.abs(constant) > 1e-13:
        pauli_strings.append("I" * hilbert.size)
        weights.append(constant)

    # the calculation above returns complex weights even for operators that are
    # purely real, so we discard their imaginary parts
    weights = [x.real if np.isreal(x) else x for x in weights]

    res = cls(hilbert, operators=pauli_strings, weights=weights, dtype=dtype)
    return res



# Copyright 2021 The NetKet Authors - All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax
import jax.numpy as jnp

from netket import config
from netket.utils.struct import dataclass, field

default_dtype = jnp.float64 if config.netket_enable_x64 else jnp.float32


@dataclass
class TableauRKExplicit:
    r"""
    Class representing the Butcher tableau of an explicit Runge-Kutta method [1,2],
    which, given the ODE :math:`dy/dt = F(t, y)`, updates the solution as

    .. math::
        y_{t+dt} = y_t + \sum_l b_l k_l

    with the intermediate slopes

    .. math::
        k_l = F(t + c_l dt, y_t + \sum_{m < l} a_{lm} k_m).

    If :code:`self.is_adaptive`, the tableau also contains the coefficients :math:`b'_l`
    which can be used to estimate the local truncation error by the formula

    .. math::
        y_{\mathrm{err}} = \sum_l (b_l - b'_l) k_l.

    [1] https://en.wikipedia.org/w/index.php?title=Runge%E2%80%93Kutta_methods&oldid=1055669759
    [2] J. Stoer and R. Bulirsch, Introduction to Numerical Analysis, Springer NY (2002).
    """

    order: tuple[int, int]
    """The order of the tableau"""

    a: jax.numpy.ndarray = field(repr=False)
    """Coefficients of th intermediate states."""
    b: jax.numpy.ndarray = field(repr=False)
    """Coefficients of the intermediate slopes."""
    c: jax.numpy.ndarray = field(repr=False)
    """Coefficients of the intermediate times."""

    name: str = field(pytree_node=False, default="RKTableau")
    """The name of the tableau."""

    def __repr__(self):
        return self.name

    @property
    def is_adaptive(self):
        """Boolean indication whether the integrator can beå adaptive."""
        return self.b.ndim == 2


# fmt: off
# flake8: noqa: E123, E126, E201, E202, E221, E226, E231, E241, E251

# Fixed Step methods
bt_feuler = TableauRKExplicit(
                order = (1,),
                a = jnp.zeros((1,1), dtype=default_dtype),
                b = jnp.ones((1,), dtype=default_dtype),
                c = jnp.zeros((1), dtype=default_dtype),
                name = "Euler"
                )


bt_midpoint = TableauRKExplicit(
                order = (2,),
                a = jnp.array([[0,   0],
                               [1/2, 0]], dtype=default_dtype),
                b = jnp.array( [0,   1], dtype=default_dtype),
                c = jnp.array( [0, 1/2], dtype=default_dtype),
                name = "Midpoint"
                )


bt_heun = TableauRKExplicit(
                order = (2,),
                a = jnp.array([[0,   0],
                               [1,   0]], dtype=default_dtype),
                b = jnp.array( [1/2, 1/2], dtype=default_dtype),
                c = jnp.array( [0, 1], dtype=default_dtype),
                name = "Heun"
                )


bt_rk4  = TableauRKExplicit(
                order = (4,),
                a = jnp.array([[0,   0,   0,   0],
                               [1/2, 0,   0,   0],
                               [0,   1/2, 0,   0],
                               [0,   0,   1,   0]], dtype=default_dtype),
                b = jnp.array( [1/6,  1/3,  1/3,  1/6], dtype=default_dtype),
                c = jnp.array( [0, 1/2, 1/2, 1], dtype=default_dtype),
                name = "RK4"
                )


# Adaptive step:
# Heun Euler https://en.wikipedia.org/wiki/Runge–Kutta_methods
bt_rk12  = TableauRKExplicit(
                order = (2,1),
                a = jnp.array([[0,   0],
                               [1,   0]], dtype=default_dtype),
                b = jnp.array([[1/2, 1/2],
                               [1,   0]], dtype=default_dtype),
                c = jnp.array( [0, 1], dtype=default_dtype),
                name = "RK12"
                )


# Bogacki–Shampine coefficients
bt_rk23  = TableauRKExplicit(
                order = (3,2),
                a = jnp.array([[0,   0,   0,   0],
                               [1/2, 0,   0,   0],
                               [0,   3/4, 0,   0],
                               [2/9, 1/3, 4/9, 0]], dtype=default_dtype),
                b = jnp.array([[7/24,1/4, 1/3, 1/8],
                               [2/9, 1/3, 4/9, 0]], dtype=default_dtype),
                c = jnp.array( [0, 1/2, 3/4, 1], dtype=default_dtype),
                name = "RK23"
                )


bt_rk4_fehlberg = TableauRKExplicit(
                order = (5,4),
                a = jnp.array([[ 0,          0,          0,           0,            0,      0 ],
                              [  1/4,        0,          0,           0,            0,      0 ],
                              [  3/32,       9/32,       0,           0,            0,      0 ],
                              [  1932/2197,  -7200/2197, 7296/2197,   0,            0,      0 ],
                              [  439/216,    -8,         3680/513,    -845/4104,    0,      0 ],
                              [  -8/27,      2,          -3544/2565,  1859/4104,    11/40,  0 ]], dtype=default_dtype),
                b = jnp.array([[ 25/216,     0,          1408/2565,   2197/4104,    -1/5,   0 ],
                               [ 16/135,     0,          6656/12825,  28561/56430,  -9/50,  2/55]], dtype=default_dtype),
                c = jnp.array( [  0,         1/4,        3/8,         12/13,        1,      1/2], dtype=default_dtype),
                name = "RK45Fehlberg"
                )


bt_rk4_dopri  = TableauRKExplicit(
                order = (5,4),
                a = jnp.array([[ 0,           0,           0,           0,        0,             0,         0 ],
                              [  1/5,         0,           0,           0,        0,             0,         0 ],
                              [  3/40,        9/40,        0,           0,        0,             0,         0 ],
                              [  44/45,       -56/15,      32/9,        0,        0,             0,         0 ],
                              [  19372/6561,  -25360/2187, 64448/6561,  -212/729, 0,             0,         0 ],
                              [  9017/3168,   -355/33,     46732/5247,  49/176,   -5103/18656,   0,         0 ],
                              [  35/384,      0,           500/1113,    125/192,  -2187/6784,    11/84,     0 ]], dtype=default_dtype),
                b = jnp.array([[ 35/384,      0,           500/1113,    125/192,  -2187/6784,    11/84,     0 ],
                               [ 5179/57600,  0,           7571/16695,  393/640,  -92097/339200, 187/2100,  1/40 ]], dtype=default_dtype),
                c = jnp.array( [ 0,           1/5,         3/10,        4/5,      8/9,           1,         1], dtype=default_dtype),
                name = "RK45"
                )



# Copyright 2024 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax
import jax.numpy as jnp

from netket.utils import struct
from netket.utils.types import Scalar, Array

from .base import (
    DiscreteHilbertConstraint,
)


class SumOnPartitionConstraint(DiscreteHilbertConstraint):
    """
    Constraint of an Hilbert space enforcing a total sum of all the values
    in partitions of different sizes.

    Constructed by specifying the tuples of sums on every partition and the length
    of every partition.

    This can be used to represent tensor products of Spin subsystems with
    different total magnetizations, or constraint on the number of fermions
    on different parts.
    """

    sum_values: tuple[Scalar, ...] = struct.field(pytree_node=False)
    sizes: tuple[int, ...] = struct.field(pytree_node=False, default=None)

    def __init__(self, sum_values, sizes):
        if not (isinstance(sum_values, tuple) and isinstance(sizes, tuple)):
            raise TypeError("sum_values and sizes must be tuples.")
        if not len(sum_values) == len(sizes):
            raise ValueError("Length mismatch between sum values and sizes")
        if any(v is None for v in sum_values):
            raise TypeError("None not supported as a sum constraint.")

        self.sum_values = sum_values
        self.sizes = sizes

    @jax.jit
    def __call__(self, x: Array) -> Array:
        s0 = 0
        conditions = []
        for N, sv in zip(self.sizes, self.sum_values):
            conditions.append(x[..., s0 : s0 + N].sum(axis=-1) == sv)
            s0 = s0 + N
        return jax.tree_util.tree_reduce(jnp.logical_and, conditions)

    def __hash__(self):
        return hash(("SumOnPartitionConstraint", self.sum_values, self.sizes))

    def __eq__(self, other):
        if isinstance(other, SumOnPartitionConstraint):
            return self.sum_values == other.sum_values and self.sizes == other.sizes
        return False

    def __repr__(self):
        return f"SumOnPartitionConstraint(sum_values={self.sum_values}, sizes={self.sizes})"



import flax.linen as nn
import jax.numpy as jnp

from flax.linen.dtypes import promote_dtype
from flax.linen.initializers import normal

from netket.utils.types import DType, Array, NNInitFunc


class Gaussian(nn.Module):
    r"""
    Multivariate Gaussian function with mean 0 and parametrised covariance matrix
    :math:`\Sigma_{ij}`.

    The wavefunction is given by the formula: :math:`\Psi(x) = \exp(\sum_{ij} x_i \Sigma_{ij} x_j)`.
    The (positive definite) :math:`\Sigma_{ij} = AA^T` matrix is stored as
    non-positive definite matrix A.
    """

    param_dtype: DType = jnp.float64
    """The dtype of the weights."""
    kernel_init: NNInitFunc = normal(stddev=1.0)
    """Initializer for the weights."""

    @nn.compact
    def __call__(self, x_in: Array):
        nv = x_in.shape[-1]

        kernel = self.param("kernel", self.kernel_init, (nv, nv), self.param_dtype)
        kernel = jnp.dot(kernel.T, kernel)

        kernel, x_in = promote_dtype(kernel, x_in, dtype=None)
        y = -0.5 * jnp.einsum("...i,ij,...j", x_in, kernel, x_in)

        return y



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import platform

from ._common import exec_in_terminal, version, is_available
from ._cpu_info import cpu_info
from ._mpi_info import get_global_mpi_info, get_link_flags

PLATFORM = sys.platform

STD_SPACE = 20


def printfmt(key, value=None, *, indent=0, pre="", alignment=STD_SPACE):
    INDENT_STR = "- "

    if indent > 0:
        indent_spaces = indent * 2
        alignment -= indent_spaces + len(INDENT_STR)

        pre = f"{pre}{'' : <{indent_spaces}}{INDENT_STR}"

    if value is not None:
        print(f"{pre}{key : <{alignment}} : {value}", flush=True)
    else:
        print(f"{pre}{key : <{alignment}}", flush=True)


def _fmt_device(dev):
    return f"<{dev.id}: {dev.device_kind}>"


def info():
    """
    When called via::

        # python3 -m netket.tools.check_mpi
        mpi4py_available     : True
        mpi4jax_available : True
        n_nodes           : 1

    this will print out basic MPI information to make allow users to check whether
    the environment has been set up correctly.
    """
    print("====================================================")
    print("==         NetKet Diagnostic Information          ==")
    print("====================================================")

    # try to import version without import netket itself
    from .. import _version

    printfmt("NetKet version", _version.version)
    print()

    print("# Python")
    printfmt("implementation", platform.python_implementation(), indent=1)
    printfmt("version", platform.python_version(), indent=1)
    printfmt("distribution", platform.python_compiler(), indent=1)
    printfmt("path", sys.executable, indent=1)
    print()

    # Try to detect platform
    print("# Host information")
    printfmt("System      ", platform.platform(), indent=1)
    printfmt("Architecture", platform.machine(), indent=1)

    # Try to query cpu info
    platform_info = cpu_info()

    printfmt("AVX", platform_info["supports_avx"], indent=1)
    printfmt("AVX2", platform_info["supports_avx2"], indent=1)
    printfmt("Cores", platform_info["core_count"], indent=1)
    print()

    # try to load jax
    print("# NetKet dependencies")
    printfmt("numpy", version("numpy"), indent=1)
    printfmt("jaxlib", version("jaxlib"), indent=1)
    printfmt("jax", version("jax"), indent=1)
    printfmt("equinox", version("equinox"), indent=1)
    printfmt("flax", version("flax"), indent=1)
    printfmt("optax", version("optax"), indent=1)
    printfmt("numba", version("numba"), indent=1)
    printfmt("mpi4py", version("mpi4py"), indent=1)
    printfmt("mpi4jax", version("mpi4jax"), indent=1)
    printfmt("netket", version("netket"), indent=1)
    print()

    if is_available("jax"):
        print("# Jax ")
        import jax

        backends = _jax_backends()
        printfmt("backends", backends, indent=1)
        for backend in backends:
            printfmt(
                f"{backend}",
                [_fmt_device(dev) for dev in jax.devices(backend)],
                indent=2,
            )
        print()

    if is_available("mpi4jax"):
        print("# MPI4JAX")
        import mpi4jax

        if hasattr(mpi4jax, "has_cuda_support"):
            printfmt("HAS_GPU_EXT", mpi4jax.has_cuda_support(), indent=1)
        elif hasattr(mpi4jax, "_src"):
            if hasattr(mpi4jax._src, "xla_bridge"):
                if hasattr(mpi4jax._src.xla_bridge, "HAS_GPU_EXT"):
                    printfmt(
                        "HAS_GPU_EXT", mpi4jax._src.xla_bridge.HAS_GPU_EXT, indent=1
                    )
        print()

    if is_available("mpi4py"):
        print("# MPI ")
        import mpi4py
        from mpi4py import MPI

        printfmt("mpi4py", indent=1)
        mpi4py_config = mpi4py.get_config()
        if "mpicc" in mpi4py_config:
            printfmt("MPICC", mpi4py_config["mpicc"], indent=1)
            printfmt(
                "MPI link flags",
                get_link_flags(exec_in_terminal([mpi4py_config["mpicc"], "-show"])),
                indent=1,
            )
        printfmt("MPI version", MPI.Get_version(), indent=2)
        printfmt("MPI library_version", MPI.Get_library_version(), indent=2)

        global_info = get_global_mpi_info()
        printfmt("global", indent=1)
        printfmt("MPICC", global_info["mpicc"], indent=2)
        printfmt("MPI link flags", global_info["link_flags"], indent=2)
        print()


def _jax_backends():
    import jax

    backends = []
    for backend in ["cpu", "gpu", "tpu"]:
        try:
            jax.devices(backend)
        except RuntimeError:
            pass
        else:
            backends.append(backend)
    return backends


info()
